import random
from itertools import chain

import pytest
import torch
from botorch.acquisition import (
    qExpectedImprovement,
    qLogExpectedImprovement,
    qLogNoisyExpectedImprovement,
    qNoisyExpectedImprovement,
)
from botorch.acquisition.objective import GenericMCObjective
from pydantic import ValidationError

import bofire.data_models.strategies.api as data_models
import bofire.data_models.surrogates.api as surrogate_data_models
from bofire.benchmarks.multi import C2DTLZ2, DTLZ2, CrossCoupling
from bofire.data_models.acquisition_functions.api import (
    qEI,
    qLogEI,
    qLogNEI,
    qNEI,
)
from bofire.data_models.domain.api import Outputs
from bofire.data_models.strategies.api import (
    PolytopeSampler as PolytopeSamplerDataModel,
)
from bofire.strategies.api import PolytopeSampler, QparegoStrategy
from tests.bofire.strategies.test_base import domains
from tests.bofire.utils.test_multiobjective import invalid_domains

VALID_BOTORCH_QPAREGO_STRATEGY_SPEC = {
    "domain": domains[6],
    "surrogate_specs": surrogate_data_models.BotorchSurrogates(
        surrogates=[
            surrogate_data_models.SingleTaskGPSurrogate(
                inputs=domains[6].inputs,
                outputs=Outputs(features=[domains[6].outputs.get_by_key("of1")]),
            ),
            surrogate_data_models.SingleTaskGPSurrogate(
                inputs=domains[6].inputs,
                outputs=Outputs(features=[domains[6].outputs.get_by_key("of2")]),
            ),
        ]
    ),
    "descriptor_method": "FREE",
    # "acquisition_function": specs.acquisition_functions.valid().obj(),
    "categorical_method": "FREE",
}

# BotorchSurrogates(
#                models=[
#                    SingleTaskGPSurrogate(
#                        inputs=domains[1].inputs,
#                        outputs=domains[1].outputs,
#                        input_preprocessing_specs={
#                            "if5": CategoricalEncodingEnum.ONE_HOT,
#                            "if6": CategoricalEncodingEnum.ONE_HOT,
#                        },
#                    )
#                ]
#            ),

BOTORCH_QPAREGO_STRATEGY_SPECS = {
    "valids": [
        VALID_BOTORCH_QPAREGO_STRATEGY_SPEC,
        {
            **VALID_BOTORCH_QPAREGO_STRATEGY_SPEC,
            "domain": domains[2],
            "surrogate_specs": surrogate_data_models.BotorchSurrogates(
                surrogates=[
                    surrogate_data_models.MixedSingleTaskGPSurrogate(
                        inputs=domains[2].inputs,
                        outputs=Outputs(
                            features=[domains[2].outputs.get_by_key("of1")]
                        ),
                    ),
                    surrogate_data_models.MixedSingleTaskGPSurrogate(
                        inputs=domains[2].inputs,
                        outputs=Outputs(
                            features=[domains[2].outputs.get_by_key("of2")]
                        ),
                    ),
                ]
            ),
            "descriptor_method": "EXHAUSTIVE",
            "categorical_method": "EXHAUSTIVE",
        },
        {**VALID_BOTORCH_QPAREGO_STRATEGY_SPEC, "seed": 1},
        # {**VALID_BOTORCH_QPAREGO_STRATEGY_SPEC, "surrogate_specs": VALID_MODEL_SPEC_LIST},
    ],
    "invalids": [
        {**VALID_BOTORCH_QPAREGO_STRATEGY_SPEC, "descriptor_method": None},
        {**VALID_BOTORCH_QPAREGO_STRATEGY_SPEC, "categorical_method": None},
        # {**VALID_BOTORCH_QPAREGO_STRATEGY_SPEC, "seed": -1},
    ],
}


@pytest.mark.parametrize(
    "domain",
    [
        invalid_domains[0],
        invalid_domains[1],
    ],
)
def test_invalid_qparego_init_domain(domain):
    with pytest.raises(ValidationError):
        data_models.QparegoStrategy(domain=domain)


@pytest.mark.parametrize(
    "num_test_candidates",
    list(range(1, 2)),
)
def test_qparego(num_test_candidates):
    # generate data
    benchmark = DTLZ2(dim=6)
    random_strategy = PolytopeSampler(
        data_model=PolytopeSamplerDataModel(domain=benchmark.domain)
    )
    experiments = benchmark.f(random_strategy._ask(n=10), return_complete=True)
    # init strategy
    acqfs = [qEI(), qLogEI(), qLogNEI(), qNEI()]
    b_acqfs = [
        qExpectedImprovement,
        qLogExpectedImprovement,
        qLogNoisyExpectedImprovement,
        qNoisyExpectedImprovement,
    ]
    i = random.choice([0, 1, 2, 3])

    data_model = data_models.QparegoStrategy(
        domain=benchmark.domain, acquisition_function=acqfs[i]
    )
    my_strategy = QparegoStrategy(data_model=data_model)
    my_strategy.tell(experiments)
    # test get objective
    objective, _, _ = my_strategy._get_objective_and_constraints()
    assert isinstance(objective, GenericMCObjective)
    acqfs = my_strategy._get_acqfs(2)
    assert len(acqfs) == 2
    assert isinstance(acqfs[0], b_acqfs[i])
    assert isinstance(acqfs[1], b_acqfs[i])
    # ask
    candidates = my_strategy.ask(num_test_candidates)
    assert len(candidates) == num_test_candidates


@pytest.mark.parametrize(
    "num_test_candidates",
    list(range(1, 2)),
)
def test_qparego_constraints(num_test_candidates):
    # generate data
    benchmark = C2DTLZ2(dim=4)
    random_strategy = PolytopeSampler(
        data_model=PolytopeSamplerDataModel(domain=benchmark.domain)
    )
    experiments = benchmark.f(random_strategy._ask(n=10), return_complete=True)
    # init strategy
    data_model = data_models.QparegoStrategy(domain=benchmark.domain)
    my_strategy = QparegoStrategy(data_model=data_model)
    my_strategy.tell(experiments)
    # test get objective
    objective, _, _ = my_strategy._get_objective_and_constraints()
    assert isinstance(objective, GenericMCObjective)
    # ask
    candidates = my_strategy.ask(num_test_candidates)
    assert len(candidates) == num_test_candidates


@pytest.mark.parametrize(
    "specs, benchmark, num_experiments, num_candidates",
    [
        (
            BOTORCH_QPAREGO_STRATEGY_SPECS["valids"][0],
            DTLZ2(dim=6),
            random.randint(8, 10),
            random.randint(1, 3),
        ),
        (
            BOTORCH_QPAREGO_STRATEGY_SPECS["valids"][1],
            CrossCoupling(),
            random.randint(8, 10),
            random.randint(1, 3),
        ),
    ],
)
@pytest.mark.slow
def test_get_acqf_input(specs, benchmark, num_experiments, num_candidates):
    # generate data
    random_strategy = PolytopeSampler(
        data_model=PolytopeSamplerDataModel(domain=benchmark.domain)
    )
    experiments = benchmark.f(
        random_strategy._ask(n=num_experiments), return_complete=True
    )
    print(specs.items())
    data_model = data_models.QparegoStrategy(
        domain=benchmark.domain,
        **{
            key: value
            for key, value in specs.items()
            if key not in ["domain", "surrogate_specs"]
        }
    )
    strategy = QparegoStrategy(data_model=data_model)
    # just to ensure there are no former experiments/ candidates already stored in the domain

    strategy.tell(experiments)
    strategy.ask(candidate_count=num_candidates, add_pending=True)

    X_train, X_pending = strategy.get_acqf_input_tensors()

    _, names = strategy.domain.inputs._get_transform_info(
        specs=strategy.surrogate_specs.input_preprocessing_specs
    )

    assert torch.is_tensor(X_train)
    assert torch.is_tensor(X_pending)
    assert X_train.shape == (
        num_experiments,
        len(set(chain(*names.values()))),
    )
    assert X_pending.shape == (
        num_candidates,
        len(set(chain(*names.values()))),
    )



from typing import Literal

from pydantic import Field

from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.kernels.api import (
    PolynomialKernel,
)
from bofire.data_models.priors.api import BOTORCH_NOISE_PRIOR, AnyPrior

# from bofire.data_models.strategies.api import FactorialStrategy
from bofire.data_models.surrogates.botorch import BotorchSurrogate
from bofire.data_models.surrogates.scaler import ScalerEnum
from bofire.data_models.surrogates.trainable import TrainableSurrogate


class PolynomialSurrogate(BotorchSurrogate, TrainableSurrogate):
    type: Literal["PolynomialSurrogate"] = "PolynomialSurrogate"

    kernel: PolynomialKernel = Field(default_factory=lambda: PolynomialKernel(power=2))
    noise_prior: AnyPrior = Field(default_factory=lambda: BOTORCH_NOISE_PRIOR())
    scaler: ScalerEnum = ScalerEnum.NORMALIZE

    @staticmethod
    def from_power(power: int, inputs: Inputs, outputs: Outputs):
        return PolynomialSurrogate(
            kernel=PolynomialKernel(power=power), inputs=inputs, outputs=outputs
        )



from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from bofire.data_models.features.feature import Input, TTransform


class NumericalInput(Input):
    """Abstract base class for all numerical (ordinal) input features."""

    unit: Optional[str] = None

    @staticmethod
    def valid_transform_types() -> List:
        return []

    def to_unit_range(
        self, values: Union[pd.Series, np.ndarray], use_real_bounds: bool = False
    ) -> Union[pd.Series, np.ndarray]:
        """Convert to the unit range between 0 and 1.

        Args:
            values (pd.Series): values to be transformed
            use_real_bounds (bool, optional): if True, use the bounds from the actual values else the bounds from the feature.
                Defaults to False.

        Raises:
            ValueError: If lower_bound == upper bound an error is raised

        Returns:
            pd.Series: transformed values.
        """
        if use_real_bounds:
            lower, upper = self.get_bounds(transform_type=None, values=values)
            lower = lower[0]
            upper = upper[0]
        else:
            lower, upper = self.lower_bound, self.upper_bound  # type: ignore
        if lower == upper:
            raise ValueError("Fixed feature cannot be transformed to unit range.")
        valrange = upper - lower
        return (values - lower) / valrange

    def from_unit_range(
        self, values: Union[pd.Series, np.ndarray]
    ) -> Union[pd.Series, np.ndarray]:
        """Convert from unit range.

        Args:
            values (pd.Series): values to transform from.

        Raises:
            ValueError: if the feature is fixed raise a value error.

        Returns:
            pd.Series: _description_
        """
        if self.is_fixed():
            raise ValueError("Fixed feature cannot be transformed from unit range.")
        valrange = self.upper_bound - self.lower_bound  # type: ignore
        return (values * valrange) + self.lower_bound  # type: ignore

    def is_fixed(self):
        """Method to check if the feature is fixed

        Returns:
            Boolean: True when the feature is fixed, false otherwise.
        """
        return self.lower_bound == self.upper_bound  # type: ignore

    def fixed_value(
        self, transform_type: Optional[TTransform] = None
    ) -> Union[None, List[float]]:
        """Method to get the value to which the feature is fixed

        Returns:
            Float: Return the feature value or None if the feature is not fixed.
        """
        assert transform_type is None
        if self.is_fixed():
            return [self.lower_bound]  # type: ignore
        else:
            return None

    def validate_experimental(self, values: pd.Series, strict=False) -> pd.Series:
        """Method to validate the experimental dataFrame

        Args:
            values (pd.Series): A dataFrame with experiments
            strict (bool, optional): Boolean to distinguish if the occurence of fixed features in the dataset should be considered or not.
                Defaults to False.

        Raises:
            ValueError: when a value is not numerical
            ValueError: when there is no variation in a feature provided by the experimental data

        Returns:
            pd.Series: A dataFrame with experiments
        """
        try:
            values = pd.to_numeric(values, errors="raise").astype("float64")
        except ValueError:
            raise ValueError(
                f"not all values of input feature `{self.key}` are numerical"
            )
        values = values.astype("float64")
        if strict:
            lower, upper = self.get_bounds(transform_type=None, values=values)
            if lower == upper:
                raise ValueError(
                    f"No variation present or planned for feature {self.key}. Remove it."
                )
        return values

    def validate_candidental(self, values: pd.Series) -> pd.Series:
        """Validate the suggested candidates for the feature.

        Args:
            values (pd.Series): suggested candidates for the feature

        Raises:
            ValueError: Error is raised when one of the values is not numerical.

        Returns:
            pd.Series: the original provided candidates
        """
        try:
            values = pd.to_numeric(values, errors="raise").astype("float64")
        except ValueError:
            raise ValueError(
                f"not all values of input feature `{self.key}` are numerical"
            )
        return values

    def get_bounds(
        self,
        transform_type: Optional[TTransform] = None,
        values: Optional[pd.Series] = None,
    ) -> Tuple[List[float], List[float]]:
        assert transform_type is None
        if values is None:
            return [self.lower_bound], [self.upper_bound]  # type: ignore
        lower = min(self.lower_bound, values.min())  # type: ignore
        upper = max(self.upper_bound, values.max())  # type: ignore
        return [lower], [upper]  # type: ignore



from typing import Literal

from bofire.data_models.kernels.kernel import Kernel


class MolecularKernel(Kernel):
    pass


class TanimotoKernel(MolecularKernel):
    type: Literal["TanimotoKernel"] = "TanimotoKernel"
    ard: bool = True



from abc import abstractmethod
from typing import Annotated, Optional, Type

from pydantic import Field, validator

from bofire.data_models.base import BaseModel
from bofire.data_models.constraints.api import Constraint
from bofire.data_models.domain.api import Domain
from bofire.data_models.features.api import Feature


class Strategy(BaseModel):
    type: str
    domain: Domain
    seed: Optional[Annotated[int, Field(ge=0)]] = None

    @validator("domain")
    def validate_constraints(cls, domain: Domain):
        """Validator to ensure that all constraints defined in the domain are valid for the chosen strategy

        Args:
            domain (Domain): The domain to be used in the strategy

        Raises:
            ValueError: if a constraint is defined in the domain but is invalid for the strategy chosen

        Returns:
            Domain: the domain
        """
        for constraint in domain.constraints:
            if not cls.is_constraint_implemented(type(constraint)):
                raise ValueError(
                    f"constraint `{type(constraint)}` is not implemented for strategy `{cls.__name__}`"  # type: ignore
                )
        return domain

    @validator("domain")
    def validate_features(cls, domain: Domain):
        """Validator to ensure that all features defined in the domain are valid for the chosen strategy

        Args:
            domain (Domain): The domain to be used in the strategy

        Raises:
            ValueError: if a feature type is defined in the domain but is invalid for the strategy chosen

        Returns:
            Domain: the domain
        """
        for feature in domain.inputs + domain.outputs:
            if not cls.is_feature_implemented(type(feature)):
                raise ValueError(
                    f"feature `{type(feature)}` is not implemented for strategy `{cls.__name__}`"  # type: ignore
                )
        return domain

    @validator("domain")
    def validate_input_feature_count(cls, domain: Domain):
        """Validator to ensure that at least one input is defined.

        Args:
            domain (Domain): The domain to be used in the strategy

        Raises:
            ValueError: if no input feature is specified

        Returns:
            Domain: the domain
        """
        if len(domain.inputs) == 0:
            raise ValueError("no input feature specified")
        return domain

    @classmethod
    @abstractmethod
    def is_constraint_implemented(cls, my_type: Type[Constraint]) -> bool:
        """Abstract method to check if a specific constraint type is implemented for the strategy

        Args:
            my_type (Type[Constraint]): Constraint class

        Returns:
            bool: True if the constraint type is valid for the strategy chosen, False otherwise
        """
        pass

    @classmethod
    @abstractmethod
    def is_feature_implemented(cls, my_type: Type[Feature]) -> bool:
        """Abstract method to check if a specific feature type is implemented for the strategy

        Args:
            my_type (Type[Feature]): Feature class

        Returns:
            bool: True if the feature type is valid for the strategy chosen, False otherwise
        """
        pass



import warnings
from typing import Optional, Tuple

import pandas as pd

import bofire.strategies.api as strategies
import bofire.strategies.mapper as strategy_mapper
from bofire.benchmarks.api import Hyperopt
from bofire.data_models.domain.api import Domain
from bofire.data_models.enum import RegressionMetricsEnum
from bofire.data_models.objectives.api import MinimizeObjective
from bofire.data_models.strategies.api import (
    FactorialStrategy,
    RandomStrategy,
    SoboStrategy,
)
from bofire.data_models.surrogates.api import AnyTrainableSurrogate
from bofire.runners.run import run


def hyperoptimize(
    surrogate_data: AnyTrainableSurrogate,
    training_data: pd.DataFrame,
    folds: int,
    random_state: Optional[int] = None,
) -> Tuple[AnyTrainableSurrogate, pd.DataFrame]:
    if surrogate_data.hyperconfig is None:
        warnings.warn(
            "No hyperopt is possible as no hyperopt config is available. Returning initial config."
        )
        return surrogate_data, pd.DataFrame({e.name: [] for e in RegressionMetricsEnum})

    def best(domain: Domain, experiments: pd.DataFrame) -> float:
        return (
            experiments[domain.outputs[0].key].min()
            if isinstance(domain.outputs[0].objective, MinimizeObjective)
            else experiments[domain.outputs[0].key].max()
        )

    def sample(domain):
        datamodel = RandomStrategy(domain=domain)
        sampler = strategies.map(data_model=datamodel)
        sampled = sampler.ask(len(domain.inputs) + 1)
        return sampled

    benchmark = Hyperopt(
        surrogate_data=surrogate_data,
        training_data=training_data,
        folds=folds,
        random_state=random_state,
    )

    if surrogate_data.hyperconfig.hyperstrategy == "FactorialStrategy":  # type: ignore
        strategy = strategies.map(FactorialStrategy(domain=benchmark.domain))
        experiments = benchmark.f(
            strategy.ask(candidate_count=None), return_complete=True
        )
    else:
        strategy_data = (
            RandomStrategy
            if surrogate_data.hyperconfig.hyperstrategy == "RandomStrategy"
            else SoboStrategy
        )
        experiments = run(
            benchmark=benchmark,
            strategy_factory=lambda domain: strategy_mapper.map(
                data_model=strategy_data(domain=domain)
            ),
            metric=best,
            n_runs=1,
            n_iterations=surrogate_data.hyperconfig.n_iterations  # type: ignore
            - len(benchmark.domain.inputs)
            - 1,
            initial_sampler=sample,
            n_procs=1,
        )[0][0]

    # analyze the results and get the best
    experiments = experiments.sort_values(
        by=benchmark.target_metric.name,
        ascending=True
        if isinstance(benchmark.domain.outputs[0].objective, MinimizeObjective)
        else False,
    )

    surrogate_data.update_hyperparameters(experiments.iloc[0])

    return (
        surrogate_data,
        experiments[
            surrogate_data.hyperconfig.domain.inputs.get_keys()
            + [e.name for e in RegressionMetricsEnum]
        ],
    )



import random
import uuid

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

import tests.bofire.data_models.specs.api as specs
from bofire.data_models.constraints.api import (
    ConstraintNotFulfilledError,
    LinearEqualityConstraint,
)
from bofire.data_models.domain.api import Domain
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
    DiscreteInput,
    Input,
    Output,
)

if1 = specs.features.valid(ContinuousInput).obj(key="cont")
if2 = specs.features.valid(CategoricalInput).obj(key="cat")
if3 = specs.features.valid(CategoricalDescriptorInput).obj(key="cat_")
if4 = specs.features.valid(CategoricalInput).obj(
    key="cat2", allowed=[True, True, False]
)
if5 = specs.features.valid(ContinuousInput).obj(
    key="if5",
    bounds=(3, 3),
)
if6 = specs.features.valid(CategoricalInput).obj(
    key="if6", categories=["c1", "c2", "c3"], allowed=[True, False, False]
)
of1 = specs.features.valid(ContinuousOutput).obj(key="out1")
of2 = specs.features.valid(ContinuousOutput).obj(key="out2")


def generate_experiments(
    domain: Domain,
    row_count: int = 5,
    tol: float = 0.0,
    only_allowed_categories: bool = False,
    force_all_categories: bool = False,
    include_labcode: bool = False,
):
    if only_allowed_categories:
        assert force_all_categories is False
    if force_all_categories:
        assert only_allowed_categories is False
    experiments = pd.DataFrame.from_dict(
        [
            {
                **{
                    f.key: random.uniform(f.lower_bound - tol, f.upper_bound + tol)  # type: ignore
                    for f in domain.get_features(ContinuousInput)
                },
                **{
                    f.key: random.choice(f.values)  # type: ignore
                    for f in domain.get_features(DiscreteInput)
                },
                **{
                    k: random.random()
                    for k in [
                        *domain.get_feature_keys(ContinuousOutput),
                    ]
                },
                **{
                    f.key: random.choice(f.categories)  # type: ignore
                    if not only_allowed_categories
                    else random.choice(f.get_allowed_categories())  # type: ignore
                    for f in domain.get_features(CategoricalInput)
                },
            }
            for _ in range(row_count)
        ]
    )
    if include_labcode:
        experiments["labcode"] = [str(i) for i in range(row_count)]
    if force_all_categories:
        for feat in domain.get_features(CategoricalInput):
            categories = (
                feat.categories  # type: ignore
                if len(feat.categories) <= row_count  # type: ignore
                else feat.categories[:row_count]  # type: ignore
            )
            experiments.loc[: len(categories) - 1, feat.key] = categories

    return experiments


def generate_candidates(domain: Domain, row_count: int = 5):
    return pd.DataFrame.from_dict(
        [
            {
                **{
                    feat.key: random.uniform(feat.lower_bound, feat.upper_bound)  # type: ignore
                    for feat in domain.get_features(ContinuousInput)
                },
                **{
                    f"{k}_pred": random.random()
                    for k in domain.get_feature_keys(ContinuousOutput)
                },
                **{
                    f"{k}_sd": random.random()
                    for k in domain.get_feature_keys(ContinuousOutput)
                },
                **{
                    f"{k}_des": random.random()
                    for k in domain.get_feature_keys(ContinuousOutput)
                },
                **{
                    f.key: random.choice(f.get_allowed_categories())  # type: ignore
                    for f in domain.get_features(CategoricalInput)
                },
            }
            for _ in range(row_count)
        ]
    )


def generate_invalid_candidates_bounds(domain, row_count: int = 5, error="lower"):
    candidates = generate_candidates(domain, row_count)
    if error == "lower":
        candidates.loc[0, domain.get_feature_keys(ContinuousInput)[0]] = (
            domain.get_features(ContinuousInput)[0].lower_bound - 0.1
        )
    else:
        candidates.loc[0, domain.get_feature_keys(ContinuousInput)[0]] = (
            domain.get_features(ContinuousInput)[0].upper_bound + 0.1
        )
    return candidates


domain0 = Domain.from_lists(
    inputs=[if1, if2, if3],
    outputs=[of1, of2],
    constraints=[],
)

domain1 = Domain.from_lists(
    inputs=[if1, if2],
    outputs=[of1],
    constraints=[],
)

domain2 = Domain.from_lists(
    inputs=[if1, if2, if3],
    outputs=[of1],
    constraints=[],
)

domain3 = Domain.from_lists(
    [if1, if2],
    [of1, of2],
    [],
)

domain4 = Domain.from_lists(
    [if1, if2, if3, if4],
    [of1, of2],
    [],
)
domain5 = Domain.from_lists(
    [if1, if5],
    [of1, of2],
    [],
)
domain6 = Domain.from_lists(
    [if1, if6],
    [of1, of2],
)
domain7 = Domain.from_lists(
    [if1, if5],
    [of1, of2],
    constraints=[
        LinearEqualityConstraint(features=["cont", "if5"], coefficients=[1, 1], rhs=500)
    ],
)

domains = [domain0, domain1, domain2, domain3, domain4]


@pytest.mark.parametrize(
    "domain, experiments",
    [
        (d, generate_experiments(d, include_labcode=include_labcode))
        for d in domains + [domain5, domain6]
        for include_labcode in [True, False]
        # (d, generate_experiments(d))
        # for d in domains+[domain5,domain6]
    ],
)
def test_domain_validate_experiments_valid(
    domain: Domain,
    experiments: pd.DataFrame,
):
    experiments1 = domain.validate_experiments(experiments.copy())
    for col in experiments.columns:
        experiments[col] = experiments[col].map(str)
    experiments2 = domain.validate_experiments(experiments)
    assert_frame_equal(
        left=experiments1,
        right=experiments2,
    )


@pytest.mark.parametrize(
    "domain, experiments, strict",
    [
        (domains[0], generate_experiments(domains[1]), True),
        (domains[0], generate_experiments(domains[1]), False),
    ],
)
def test_domain_validate_experiments_invalid(
    domain: Domain, experiments: pd.DataFrame, strict: bool
):
    with pytest.raises(ValueError):
        domain.validate_experiments(experiments, strict=strict)


@pytest.mark.parametrize(
    "domain, experiments",
    [
        (d, generate_experiments(d, only_allowed_categories=True))
        for d in [domain5, domain6]
    ],
)
def test_domain_validate_experiments_strict_invalid(
    domain: Domain,
    experiments: pd.DataFrame,
):
    domain.validate_experiments(experiments, strict=False)
    with pytest.raises(ValueError):
        domain.validate_experiments(experiments, strict=True)


def test_domain_validate_experiments_invalid_labcode():
    # non distinct labcodes
    experiments = generate_experiments(domain4, row_count=2, include_labcode=True)
    experiments.labcode = ["1", "1"]
    with pytest.raises(ValueError):
        domain4.validate_experiments(experiments)
    # na labcodes
    experiments = generate_experiments(domain4, row_count=2, include_labcode=True)
    experiments.labcode = [np.nan, "1"]
    with pytest.raises(ValueError):
        domain4.validate_experiments(experiments)


@pytest.mark.parametrize(
    "domain, candidates", [(d, generate_candidates(d)) for d in domains]
)
def test_domain_validate_candidates_valid(
    domain: Domain,
    candidates: pd.DataFrame,
):
    candidates1 = domain.validate_candidates(candidates.copy())
    for col in domain.inputs.get_keys():
        candidates[col] = candidates[col].map(str)
    candidates2 = domain.validate_candidates(candidates.copy())
    assert_frame_equal(
        left=candidates1,
        right=candidates2,
    )


@pytest.mark.parametrize(
    "domain, candidates",
    [
        (d, generate_candidates(d).drop(key, axis=1))
        for d in [domain0]
        for key in d.get_feature_keys(Input)
    ]
    + [
        (d, generate_candidates(d).drop(key, axis=1))
        for d in [domain0]
        for key_ in d.get_feature_keys(Output)
        for key in [f"{key_}_pred", f"{key_}_sd", f"{key_}_des"]
    ],
)
def test_domain_validate_candidates_missing_cols(
    domain: Domain,
    candidates: pd.DataFrame,
):
    with pytest.raises(ValueError):
        domain.validate_candidates(candidates)


@pytest.mark.parametrize(
    "domain, candidates",
    [
        (domain0, generate_invalid_candidates_bounds(domain0, error="lower")),
        (domain0, generate_invalid_candidates_bounds(domain0, error="upper")),
    ],
)
def test_domain_validate_candidates_invalid_bounds(
    domain: Domain, candidates: pd.DataFrame
):
    with pytest.raises(ValueError):
        domain.validate_candidates(candidates)


def test_domain_validate_candidates_invalid_categories():
    candidates = generate_candidates(domain4)
    candidates.loc[0, "cat2"] = "c3"
    with pytest.raises(ValueError):
        domain4.validate_candidates(candidates)


@pytest.mark.parametrize(
    "domain, candidates, key",
    [
        (d, generate_candidates(d), key)
        for d in domains
        for key in d.get_feature_keys(Input)
    ]
    + [
        (d, generate_candidates(d), key)
        for d in domains
        for key_ in d.get_feature_keys(ContinuousOutput)
        for key in [f"{key_}_pred", f"{key_}_sd", f"{key_}_des"]
    ],
)
def test_domain_validate_candidates_not_numerical(
    domain: Domain,
    candidates: pd.DataFrame,
    key: str,
):
    candidates[key] = str(uuid.uuid4())
    with pytest.raises(ValueError):
        domain.validate_candidates(candidates)


@pytest.mark.parametrize(
    "domain, candidates, raise_validation_error",
    [
        (d, generate_candidates(d), raise_validation_error)
        for d in [domain7]
        for raise_validation_error in [True, False]
    ],
)
def test_domain_validate_candidates_constraint_not_fulfilled(
    domain, candidates, raise_validation_error
):
    if raise_validation_error:
        with pytest.raises(ConstraintNotFulfilledError):
            domain.validate_candidates(
                candidates, raise_validation_error=raise_validation_error
            )
    else:
        assert isinstance(
            domain.validate_candidates(
                candidates, raise_validation_error=raise_validation_error
            ),
            pd.DataFrame,
        )


def test_outputs_add_valid_columns():
    experiments = generate_experiments(domain=domain0)
    assert "valid_out1" not in experiments.columns
    assert "valid_out2" not in experiments.columns
    experiments = domain0.outputs.add_valid_columns(experiments)
    assert "valid_out1" in experiments.columns
    assert "valid_out2" in experiments.columns
    #
    experiments["valid_out1"] = "1"
    experiments["valid_out2"] = "0"
    experiments = domain0.outputs.add_valid_columns(experiments)
    assert (experiments["valid_out1"] == 1).all()
    assert (experiments["valid_out2"] == 0).all()
    experiments["valid_out1"] = 0
    experiments["valid_out2"] = 1
    experiments = domain0.outputs.add_valid_columns(experiments)
    assert (experiments["valid_out1"] == 0).all()
    assert (experiments["valid_out2"] == 1).all()
    experiments["valid_out1"] = 0.0
    experiments["valid_out2"] = 1.0
    experiments = domain0.outputs.add_valid_columns(experiments)
    assert (experiments["valid_out1"] == 0).all()
    assert (experiments["valid_out2"] == 1).all()
    for _test_val in ["1.0", "1.2"]:
        experiments["valid_out1"] = _test_val
        with pytest.raises(ValueError):
            domain0.outputs.add_valid_columns(experiments)



import collections.abc
import itertools
import warnings
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
)

import numpy as np
import pandas as pd
from pydantic import Field, validator

from bofire.data_models.base import BaseModel
from bofire.data_models.constraints.api import (
    AnyConstraint,
    ConstraintNotFulfilledError,
    LinearConstraint,
    NChooseKConstraint,
)
from bofire.data_models.domain.constraints import Constraints
from bofire.data_models.domain.features import Features, Inputs, Outputs
from bofire.data_models.features.api import (
    AnyInput,
    AnyOutput,
    ContinuousInput,
    ContinuousOutput,
    Feature,
    Input,
    Output,
)
from bofire.data_models.objectives.api import Objective


def isinstance_or_union(obj, of):
    if get_origin(of) is Union:
        of = get_args(of)
    return isinstance(obj, of)


def is_numeric(s: Union[pd.Series, pd.DataFrame]) -> bool:
    if isinstance(s, pd.Series):
        return pd.to_numeric(s, errors="coerce").notnull().all()
    return s.apply(lambda s: pd.to_numeric(s, errors="coerce").notnull().all()).all()  # type: ignore


class Domain(BaseModel):
    type: Literal["Domain"] = "Domain"

    inputs: Inputs = Field(default_factory=lambda: Inputs())
    outputs: Outputs = Field(default_factory=lambda: Outputs())

    constraints: Constraints = Field(default_factory=lambda: Constraints())

    """Representation of the optimization problem/domain

    Attributes:
        inputs (List[Input], optional): List of input features. Defaults to [].
        outputs (List[Output], optional): List of output features. Defaults to [].
        constraints (List[Constraint], optional): List of constraints. Defaults to [].
    """

    @classmethod
    def from_lists(
        cls,
        inputs: Optional[Sequence[AnyInput]] = None,
        outputs: Optional[Sequence[AnyOutput]] = None,
        constraints: Optional[Sequence[AnyConstraint]] = None,
    ):
        inputs = [] if inputs is None else inputs
        outputs = [] if outputs is None else outputs
        constraints = [] if constraints is None else constraints
        return cls(
            inputs=Inputs(features=inputs),
            outputs=Outputs(features=outputs),
            constraints=Constraints(constraints=constraints),
        )

    @validator("inputs", always=True, pre=True)
    def validate_inputs_list(cls, v, values):
        if isinstance(v, collections.abc.Sequence):
            v = Inputs(features=v)
            return v
        if isinstance_or_union(v, AnyInput):
            return Inputs(features=[v])
        else:
            return v

    @validator("outputs", always=True, pre=True)
    def validate_outputs_list(cls, v, values):
        if isinstance(v, collections.abc.Sequence):
            return Outputs(features=v)
        if isinstance_or_union(v, AnyOutput):
            return Outputs(features=[v])
        else:
            return v

    @validator("constraints", always=True, pre=True)
    def validate_constraints_list(cls, v, values):
        if isinstance(v, list):
            return Constraints(constraints=v)
        if isinstance_or_union(v, AnyConstraint):
            return Constraints(constraints=[v])
        else:
            return v

    @validator("outputs", always=True)
    def validate_unique_feature_keys(cls, v: Outputs, values) -> Outputs:
        """Validates if provided input and output feature keys are unique

        Args:
            v (Outputs): List of all output features of the domain.
            value (Dict[str, Inputs]): Dict containing a list of input features as single entry.

        Raises:
            ValueError: Feature keys are not unique.

        Returns:
            Outputs: Keeps output features as given.
        """
        if "inputs" not in values:
            return v
        features = v + values["inputs"]
        keys = [f.key for f in features]
        if len(set(keys)) != len(keys):
            raise ValueError("feature keys are not unique")
        return v

    @validator("constraints", always=True)
    def validate_constraints(cls, v, values):
        """Validate if all features included in the constraints are also defined as features for the domain.

        Args:
            v (List[Constraint]): List of constraints or empty if no constraints are defined
            values (List[Input]): List of input features of the domain

        Raises:
            ValueError: Feature key in constraint is unknown.

        Returns:
            List[Constraint]: List of constraints defined for the domain
        """
        if "inputs" not in values:
            return v
        keys = [f.key for f in values["inputs"]]
        for c in v:
            if isinstance(c, LinearConstraint) or isinstance(c, NChooseKConstraint):
                for f in c.features:
                    if f not in keys:
                        raise ValueError(f"feature {f} in constraint unknown ({keys})")
        return v

    @validator("constraints", always=True)
    def validate_linear_constraints(cls, v, values):
        """Validate if all features included in linear constraints are continuous ones.

        Args:
            v (List[Constraint]): List of constraints or empty if no constraints are defined
            values (List[Input]): List of input features of the domain

        Raises:
            ValueError: _description_


        Returns:
           List[Constraint]: List of constraints defined for the domain
        """
        if "inputs" not in values:
            return v

        # gather continuous inputs in dictionary
        continuous_inputs_dict = {}
        for f in values["inputs"]:
            if isinstance(f, ContinuousInput):
                continuous_inputs_dict[f.key] = f

        # check if non continuous input features appear in linear constraints
        for c in v:
            if isinstance(c, LinearConstraint):
                for f in c.features:
                    assert f in continuous_inputs_dict, f"{f} must be continuous."
        return v

    def get_feature_reps_df(self) -> pd.DataFrame:
        """Returns a pandas dataframe describing the features contained in the optimization domain."""
        df = pd.DataFrame(
            index=self.get_feature_keys(Feature),
            columns=["Type", "Description"],
            data={
                "Type": [
                    feat.__class__.__name__ for feat in self.get_features(Feature)
                ],
                "Description": [feat.__str__() for feat in self.get_features(Feature)],
            },
        )
        return df

    def get_constraint_reps_df(self):
        """Provides a tabular overwiev of all constraints within the domain

        Returns:
            pd.DataFrame: DataFrame listing all constraints of the domain with a description
        """
        df = pd.DataFrame(
            index=range(len(self.constraints)),
            columns=["Type", "Description"],
            data={
                "Type": [feat.__class__.__name__ for feat in self.constraints],
                "Description": [
                    constraint.__str__() for constraint in self.constraints
                ],
            },
        )
        return df

    def get_features(
        self,
        includes: Union[Type[Feature], List[Type[Feature]]] = Feature,
        excludes: Union[Type[Feature], List[Type[Feature]], None] = None,
        exact: bool = False,
    ) -> Features:
        """get features of the domain

        Args:
            includes (Union[Type, List[Type]], optional): Feature class or list of specific feature classes to be returned. Defaults to Feature.
            excludes (Union[Type, List[Type]], optional): Feature class or list of specific feature classes to be excluded from the return. Defaults to None.
            exact (bool, optional): Boolean to distinguish if only the exact class listed in includes and no subclasses inherenting from this class shall be returned. Defaults to False.
            by_attribute (str, optional): If set it is filtered by the attribute specified in by `by_attribute`. Defaults to None.

        Returns:
            List[Feature]: List of features in the domain fitting to the passed requirements.
        """
        assert isinstance(self.inputs, Inputs)
        features = self.inputs + self.outputs
        return features.get(includes, excludes, exact)

    def get_feature_keys(
        self,
        includes: Union[Type, List[Type]] = Feature,
        excludes: Union[Type, List[Type]] = None,
        exact: bool = False,
    ) -> List[str]:
        """Method to get feature keys of the domain

        Args:
            includes (Union[Type, List[Type]], optional): Feature class or list of specific feature classes to be returned. Defaults to Feature.
            excludes (Union[Type, List[Type]], optional): Feature class or list of specific feature classes to be excluded from the return. Defaults to None.
            exact (bool, optional): Boolean to distinguish if only the exact class listed in includes and no subclasses inherenting from this class shall be returned. Defaults to False.

        Returns:
            List[str]: List of feature keys fitting to the passed requirements.
        """
        return [
            f.key
            for f in self.get_features(
                includes=includes,
                excludes=excludes,
                exact=exact,
            )
        ]

    def get_feature(self, key: str):
        """get a specific feature by its key

        Args:
            key (str): Feature key

        Returns:
            Feature: The feature with the passed key
        """
        assert isinstance(self.inputs, Inputs)
        return {f.key: f for f in self.inputs + self.outputs}[key]

    # TODO: tidy this up
    def get_nchoosek_combinations(self, exhaustive: bool = False):  # noqa: C901
        """get all possible NChooseK combinations

        Args:
            exhaustive (bool, optional): if True all combinations are returned. Defaults to False.

        Returns:
            Tuple(used_features_list, unused_features_list): used_features_list is a list of lists containing features used in each NChooseK combination.
                unused_features_list is a list of lists containing features unused in each NChooseK combination.
        """

        if len(self.constraints.get(NChooseKConstraint)) == 0:
            used_continuous_features = self.get_feature_keys(ContinuousInput)
            return used_continuous_features, []

        used_features_list_all = []

        # loops through each NChooseK constraint
        for con in self.constraints.get(NChooseKConstraint):
            assert isinstance(con, NChooseKConstraint)
            used_features_list = []

            if exhaustive:
                for n in range(con.min_count, con.max_count + 1):
                    used_features_list.extend(itertools.combinations(con.features, n))

                if con.none_also_valid:
                    used_features_list.append(())
            else:
                used_features_list.extend(
                    itertools.combinations(con.features, con.max_count)
                )

            used_features_list_all.append(used_features_list)

        used_features_list_all = list(
            itertools.product(*used_features_list_all)
        )  # product between NChooseK constraints

        # format into a list of used features
        used_features_list_formatted = []
        for used_features_list in used_features_list_all:
            used_features_list_flattened = [
                item for sublist in used_features_list for item in sublist
            ]
            used_features_list_formatted.append(list(set(used_features_list_flattened)))

        # sort lists
        used_features_list_sorted = []
        for used_features in used_features_list_formatted:
            used_features_list_sorted.append(sorted(used_features))

        # drop duplicates
        used_features_list_no_dup = []
        for used_features in used_features_list_sorted:
            if used_features not in used_features_list_no_dup:
                used_features_list_no_dup.append(used_features)

        # print(f"duplicates dropped: {len(used_features_list_sorted)-len(used_features_list_no_dup)}")

        # remove combinations not fulfilling constraints
        used_features_list_final = []
        for combo in used_features_list_no_dup:
            fulfil_constraints = (
                []
            )  # list of bools tracking if constraints are fulfilled
            for con in self.constraints.get(NChooseKConstraint):
                assert isinstance(con, NChooseKConstraint)
                count = 0  # count of features in combo that are in con.features
                for f in combo:
                    if f in con.features:
                        count += 1
                if count >= con.min_count and count <= con.max_count:
                    fulfil_constraints.append(True)
                elif count == 0 and con.none_also_valid:
                    fulfil_constraints.append(True)
                else:
                    fulfil_constraints.append(False)
            if np.all(fulfil_constraints):
                used_features_list_final.append(combo)

        # print(f"violators dropped: {len(used_features_list_no_dup)-len(used_features_list_final)}")

        # features unused
        features_in_cc = []
        for con in self.constraints.get(NChooseKConstraint):
            assert isinstance(con, NChooseKConstraint)
            features_in_cc.extend(con.features)
        features_in_cc = list(set(features_in_cc))
        features_in_cc.sort()
        unused_features_list = []
        for used_features in used_features_list_final:
            unused_features_list.append(
                [f_key for f_key in features_in_cc if f_key not in used_features]
            )

        # postprocess
        # used_features_list_final2 = []
        # unused_features_list2 = []
        # for used, unused in zip(used_features_list_final,unused_features_list):
        #     if len(used) == 3:
        #         used_features_list_final2.append(used), unused_features_list2.append(unused)

        return used_features_list_final, unused_features_list

    def coerce_invalids(self, experiments: pd.DataFrame) -> pd.DataFrame:
        """Coerces all invalid output measurements to np.nan

        Args:
            experiments (pd.DataFrame): Dataframe containing experimental data

        Returns:
            pd.DataFrame: coerced dataframe
        """
        # coerce invalid to nan
        for feat in self.get_feature_keys(Output):
            experiments.loc[experiments[f"valid_{feat}"] == 0, feat] = np.nan
        return experiments

    def aggregate_by_duplicates(
        self,
        experiments: pd.DataFrame,
        prec: int,
        delimiter: str = "-",
        method: Literal["mean", "median"] = "mean",
    ) -> Tuple[pd.DataFrame, list]:
        """Aggregate the dataframe by duplicate experiments

        Duplicates are identified based on the experiments with the same input features. Continuous input features
        are rounded before identifying the duplicates. Aggregation is performed by taking the average of the
        involved output features.

        Args:
            experiments (pd.DataFrame): Dataframe containing experimental data
            prec (int): Precision of the rounding of the continuous input features
            delimiter (str, optional): Delimiter used when combining the orig. labcodes to a new one. Defaults to "-".

        Returns:
            Tuple[pd.DataFrame, list]: Dataframe holding the aggregated experiments, list of lists holding the labcodes of the duplicates
        """
        # prepare the parent frame
        if method not in ["mean", "median"]:
            raise ValueError(f"Unknown aggregation type provided: {method}")

        preprocessed = self.outputs.preprocess_experiments_any_valid_output(experiments)
        assert preprocessed is not None
        experiments = preprocessed.copy()
        if "labcode" not in experiments.columns:
            experiments["labcode"] = [
                str(i + 1).zfill(int(np.ceil(np.log10(experiments.shape[0]))))
                for i in range(experiments.shape[0])
            ]

        # round it
        experiments[self.get_feature_keys(ContinuousInput)] = experiments[
            self.get_feature_keys(ContinuousInput)
        ].round(prec)

        # coerce invalid to nan
        experiments = self.coerce_invalids(experiments)

        # group and aggregate
        agg: Dict[str, Any] = {
            feat: method for feat in self.get_feature_keys(ContinuousOutput)
        }
        agg["labcode"] = lambda x: delimiter.join(sorted(x.tolist()))
        for feat in self.get_feature_keys(Output):
            agg[f"valid_{feat}"] = lambda x: 1

        grouped = experiments.groupby(self.get_feature_keys(Input))
        duplicated_labcodes = [
            sorted(group.labcode.to_numpy().tolist())
            for _, group in grouped
            if group.shape[0] > 1
        ]

        experiments = grouped.aggregate(agg).reset_index(drop=False)
        for feat in self.get_feature_keys(Output):
            experiments.loc[experiments[feat].isna(), f"valid_{feat}"] = 0

        experiments = experiments.sort_values(by="labcode")
        experiments = experiments.reset_index(drop=True)
        return experiments, sorted(duplicated_labcodes)

    def validate_experiments(
        self,
        experiments: pd.DataFrame,
        strict: bool = False,
    ) -> pd.DataFrame:
        """checks the experimental data on validity

        Args:
            experiments (pd.DataFrame): Dataframe with experimental data

        Raises:
            ValueError: empty dataframe
            ValueError: the column for a specific feature is missing the provided data
            ValueError: there are labcodes with null value
            ValueError: there are labcodes with nan value
            ValueError: labcodes are not unique
            ValueError: the provided columns do no match to the defined domain
            ValueError: the provided columns do no match to the defined domain
            ValueError: Input with null values
            ValueError: Input with nan values

        Returns:
            pd.DataFrame: The provided dataframe with experimental data
        """

        if len(experiments) == 0:
            raise ValueError("no experiments provided (empty dataframe)")
        # we allow here for a column named labcode used to identify experiments
        if "labcode" in experiments.columns:
            # test that labcodes are not na
            if experiments.labcode.isnull().to_numpy().any():
                raise ValueError("there are labcodes with null value")
            if experiments.labcode.isna().to_numpy().any():
                raise ValueError("there are labcodes with nan value")
            # test that labcodes are distinct
            if (
                len(set(experiments.labcode.to_numpy().tolist()))
                != experiments.shape[0]
            ):
                raise ValueError("labcodes are not unique")
        # run the individual validators
        experiments = self.inputs.validate_experiments(
            experiments=experiments, strict=strict
        )
        experiments = self.outputs.validate_experiments(experiments=experiments)
        return experiments

    def describe_experiments(self, experiments: pd.DataFrame) -> pd.DataFrame:
        """Method to get a tabular overview of how many measurements and how many valid entries are included in the input data for each output feature

        Args:
            experiments (pd.DataFrame): Dataframe with experimental data

        Returns:
            pd.DataFrame: Dataframe with counts how many measurements and how many valid entries are included in the input data for each output feature
        """
        data = {}
        for feat in self.get_feature_keys(Output):
            data[feat] = [
                experiments.loc[experiments[feat].notna()].shape[0],
                experiments.loc[experiments[feat].notna(), "valid_%s" % feat].sum(),
            ]
        preprocessed = self.outputs.preprocess_experiments_all_valid_outputs(
            experiments
        )
        assert preprocessed is not None
        data["all"] = [
            experiments.shape[0],
            preprocessed.shape[0],
        ]
        return pd.DataFrame.from_dict(
            data, orient="index", columns=["measured", "valid"]
        )

    def validate_candidates(
        self,
        candidates: pd.DataFrame,
        only_inputs: bool = False,
        tol: float = 1e-5,
        raise_validation_error: bool = True,
    ) -> pd.DataFrame:
        """Method to check the validty of porposed candidates

        Args:
            candidates (pd.DataFrame): Dataframe with suggested new experiments (candidates)
            only_inputs (bool,optional): If True, only the input columns are validated. Defaults to False.
            tol (float,optional): tolerance parameter for constraints. A constraint is considered as not fulfilled if the violation
                is larger than tol. Defaults to 1e-6.
            raise_validation_error (bool, optional): If true an error will be raised if candidates violate constraints,
                otherwise only a warning will be displayed. Defaults to True.

        Raises:
            ValueError: when a column is missing for a defined input feature
            ValueError: when a column is missing for a defined output feature
            ValueError: when a non-numerical value is proposed
            ValueError: when an additional column is found
            ConstraintNotFulfilledError: when the constraints are not fulfilled and `raise_validation_error = True`

        Returns:
            pd.DataFrame: dataframe with suggested experiments (candidates)
        """
        # check that each input feature has a col and is valid in itself
        assert isinstance(self.inputs, Inputs)
        candidates = self.inputs.validate_candidates(candidates)
        # check if all constraints are fulfilled
        if not self.constraints.is_fulfilled(candidates, tol=tol).all():
            if raise_validation_error:
                raise ConstraintNotFulfilledError(
                    f"Constraints not fulfilled: {candidates}"
                )
            warnings.warn("Not all constraints are fulfilled.")
        # for each continuous output feature with an attached objective object
        if not only_inputs:
            assert isinstance(self.outputs, Outputs)
            candidates = self.outputs.validate_candidates(candidates=candidates)
        return candidates

    @property
    def experiment_column_names(self):
        """the columns in the experimental dataframe

        Returns:
            List[str]: List of columns in the experiment dataframe (output feature keys + valid_output feature keys)
        """
        return self.get_feature_keys() + [
            f"valid_{output_feature_key}"
            for output_feature_key in self.get_feature_keys(Output)
        ]

    @property
    def candidate_column_names(self):
        """the columns in the candidate dataframe

        Returns:
            List[str]: List of columns in the candidate dataframe (input feature keys + input feature keys_pred, input feature keys_sd, input feature keys_des)
        """
        assert isinstance(self.outputs, Outputs)
        return (
            self.get_feature_keys(Input)
            + [
                f"{output_feature_key}_pred"
                for output_feature_key in self.outputs.get_keys_by_objective(Objective)
            ]
            + [
                f"{output_feature_key}_sd"
                for output_feature_key in self.outputs.get_keys_by_objective(Objective)
            ]
            + [
                f"{output_feature_key}_des"
                for output_feature_key in self.outputs.get_keys_by_objective(Objective)
            ]
        )

    def _set_constraints_unvalidated(
        self, constraints: Union[Sequence[AnyConstraint], Constraints]
    ):
        """Hack for reduce_domain"""
        self.constraints = Constraints(constraints=[])
        if isinstance(constraints, Constraints):
            constraints = constraints.constraints
        self.constraints.constraints = constraints



import warnings

import numpy as np
import pandas as pd

import bofire.data_models.strategies.api as data_models
from bofire.data_models.constraints.api import (
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
)
from bofire.data_models.domain.api import Domain
from bofire.data_models.features.api import (
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
    DiscreteInput,
)
from bofire.strategies.api import DoEStrategy

# from tests.bofire.strategies.botorch.test_model_spec import VALID_MODEL_SPEC_LIST

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning, append=True)

with warnings.catch_warnings():
    warnings.simplefilter("ignore")

inputs = [
    ContinuousInput(
        key=f"x{1}",
        bounds=(0.0, 1.0),
    ),
    ContinuousInput(
        key=f"x{2}",
        bounds=(0.1, 1.0),
    ),
    ContinuousInput(
        key=f"x{3}",
        bounds=(0.0, 0.6),
    ),
]
domain = Domain.from_lists(
    inputs=inputs,
    outputs=[ContinuousOutput(key="y")],
    constraints=[
        LinearEqualityConstraint(
            features=[f"x{i + 1}" for i in range(3)], coefficients=[1, 1, 1], rhs=1
        ),
        LinearInequalityConstraint(features=["x1", "x2"], coefficients=[5, 4], rhs=3.9),
        LinearInequalityConstraint(
            features=["x1", "x2"], coefficients=[-20, 5], rhs=-3
        ),
    ],
)


def test_doe_strategy_init():
    data_model = data_models.DoEStrategy(domain=domain, formula="linear")
    strategy = DoEStrategy(data_model=data_model)
    assert strategy is not None


def test_doe_strategy_ask():
    data_model = data_models.DoEStrategy(domain=domain, formula="linear")
    strategy = DoEStrategy(data_model=data_model)
    candidates = strategy.ask(candidate_count=12)
    assert candidates.shape == (12, 3)


def test_doe_strategy_ask_with_candidates():
    candidates_fixed = pd.DataFrame(
        np.array([[0.2, 0.2, 0.6], [0.3, 0.6, 0.1], [0.7, 0.1, 0.2], [0.3, 0.1, 0.6]]),
        columns=["x1", "x2", "x3"],
    )
    data_model = data_models.DoEStrategy(domain=domain, formula="linear")
    strategy = DoEStrategy(data_model=data_model)
    strategy.set_candidates(candidates_fixed)
    candidates = strategy.ask(candidate_count=12)
    assert candidates.shape == (12, 3)


def test_nchoosek_implemented():
    nchoosek_constraint = NChooseKConstraint(
        features=[f"x{i + 1}" for i in range(3)],
        min_count=0,
        max_count=2,
        none_also_valid=True,
    )
    domain = Domain.from_lists(
        inputs=[ContinuousInput(key=f"x{i + 1}", bounds=(0.0, 1.0)) for i in range(3)],
        outputs=[ContinuousOutput(key="y")],
        constraints=[nchoosek_constraint],
    )
    data_model = data_models.DoEStrategy(
        domain=domain, formula="linear", optimization_strategy="partially-random"
    )
    strategy = DoEStrategy(data_model=data_model)
    candidates = strategy.ask(candidate_count=12)
    assert candidates.shape == (12, 3)


def test_formulas_implemented():
    expected_num_candidates = {
        "linear": 7,  # 1+a+b+c+3
        "linear-and-quadratic": 10,  # 1+a+b+c+a**2+b**2+c**2+3
        "linear-and-interactions": 10,  # 1+a+b+c+ab+ac+bc+3
        "fully-quadratic": 13,  # 1+a+b+c+a**2+b**2+c**2+ab+ac+bc+3
    }

    for formula, num_candidates in expected_num_candidates.items():
        data_model = data_models.DoEStrategy(domain=domain, formula=formula)
        strategy = DoEStrategy(data_model=data_model)
        candidates = strategy.ask()
        assert candidates.shape == (num_candidates, 3)


def test_doe_strategy_correctness():
    candidates_fixed = pd.DataFrame(
        np.array([[0.2, 0.2, 0.6], [0.3, 0.6, 0.1], [0.7, 0.1, 0.2], [0.3, 0.1, 0.6]]),
        columns=["x1", "x2", "x3"],
    )
    data_model = data_models.DoEStrategy(domain=domain, formula="linear")
    strategy = DoEStrategy(data_model=data_model)
    strategy.set_candidates(candidates_fixed)
    candidates = strategy.ask(candidate_count=12)

    np.random.seed(1)
    candidates_expected = np.array(
        [[0.2, 0.2, 0.6], [0.3, 0.6, 0.1], [0.7, 0.1, 0.2], [0.3, 0.1, 0.6]]
    )
    for row in candidates.to_numpy():
        assert any(np.allclose(row, o, atol=1e-2) for o in candidates_expected)
    for o in candidates_expected[:-1]:
        assert any(np.allclose(o, row, atol=1e-2) for row in candidates.to_numpy())


def test_doe_strategy_amount_of_candidates():
    candidates_fixed = pd.DataFrame(
        np.array([[0.2, 0.2, 0.6], [0.3, 0.6, 0.1], [0.7, 0.1, 0.2], [0.3, 0.1, 0.6]]),
        columns=["x1", "x2", "x3"],
    )
    data_model = data_models.DoEStrategy(domain=domain, formula="linear")
    strategy = DoEStrategy(data_model=data_model)
    strategy.set_candidates(candidates_fixed)
    candidates = strategy.ask(candidate_count=12)

    np.random.seed(1)
    num_candidates_expected = 12
    assert len(candidates) == num_candidates_expected


def test_categorical_discrete_doe():
    quantity_a = [
        ContinuousInput(key=f"quantity_a_{i}", bounds=(0, 100)) for i in range(3)
    ]
    quantity_b = [
        ContinuousInput(key=f"quantity_b_{i}", bounds=(0, 15)) for i in range(3)
    ]
    all_inputs = [
        CategoricalInput(key="animals", categories=["Whale", "Turtle", "Sloth"]),
        DiscreteInput(key="discrete", values=[0.1, 0.2, 0.3, 1.6, 2]),
        ContinuousInput(key="independent", bounds=(3, 10)),
    ]
    all_inputs.extend(quantity_a)
    all_inputs.extend(quantity_b)

    all_constraints = [
        NChooseKConstraint(
            features=[var.key for var in quantity_a],
            min_count=0,
            max_count=1,
            none_also_valid=True,
        ),
        NChooseKConstraint(
            features=[var.key for var in quantity_b],
            min_count=0,
            max_count=2,
            none_also_valid=True,
        ),
        LinearEqualityConstraint(
            features=[var.key for var in quantity_b],
            coefficients=[1 for var in quantity_b],
            rhs=15,
        ),
    ]

    n_experiments = 10
    domain = Domain(
        inputs=all_inputs,
        outputs=[ContinuousOutput(key="y")],
        constraints=all_constraints,
    )

    data_model = data_models.DoEStrategy(
        domain=domain, formula="linear", optimization_strategy="partially-random"
    )
    strategy = DoEStrategy(data_model=data_model)
    candidates = strategy.ask(candidate_count=n_experiments)

    assert candidates.shape == (10, 9)


def test_partially_fixed_experiments():
    continuous_var = [
        ContinuousInput(key=f"continuous_var_{i}", bounds=(100, 230)) for i in range(2)
    ]

    all_constraints = [
        NChooseKConstraint(
            features=[var.key for var in continuous_var],
            min_count=1,
            max_count=2,
            none_also_valid=True,
        ),
    ]
    all_inputs = [
        CategoricalInput(key="animal", categories=["dog", "whale", "cat"]),
        CategoricalInput(key="plant", categories=["tulip", "sunflower"]),
        DiscreteInput(key="a_discrete", values=[0.1, 0.2, 0.3, 1.6, 2]),
        DiscreteInput(key="b_discrete", values=[0.1, 0.2, 0.3, 1.6, 2]),
    ]
    n_experiments = 10

    all_inputs = all_inputs + continuous_var
    domain = Domain(
        inputs=all_inputs,
        outputs=[ContinuousOutput(key="y")],
        constraints=all_constraints,
    )

    data_model = data_models.DoEStrategy(
        domain=domain,
        formula="linear",
        optimization_strategy="relaxed",
        verbose=True,
    )
    strategy = DoEStrategy(data_model=data_model)
    strategy.set_candidates(
        pd.DataFrame(
            [
                [150, 100, 0.3, 0.2, None, None],
                [0, 100, 0.3, 0.2, None, "tulip"],
                [0, 100, None, 0.2, "dog", None],
                [0, 100, 0.3, 0.2, "cat", "tulip"],
                [None, 100, 0.3, None, None, None],
            ],
            columns=[
                "continuous_var_0",
                "continuous_var_1",
                "a_discrete",
                "b_discrete",
                "animal",
                "plant",
            ],
        )
    )

    only_partially_fixed = pd.DataFrame(
        [
            [150, 100, 0.3, 0.2, None, None],
            [0, 100, 0.3, 0.2, None, "tulip"],
            [0, 100, None, 0.2, "dog", None],
            [None, 100, 0.3, None, None, None],
        ],
        columns=[
            "continuous_var_0",
            "continuous_var_1",
            "a_discrete",
            "b_discrete",
            "animal",
            "plant",
        ],
    )

    candidates = strategy.ask(candidate_count=n_experiments)
    print(candidates)
    only_partially_fixed = only_partially_fixed.mask(
        only_partially_fixed.isnull(), candidates[:4]
    )
    test_df = pd.DataFrame(np.ones((4, 6)))
    test_df = test_df.where(candidates[:4] == only_partially_fixed, 0)
    assert test_df.sum().sum() == 0


def test_categorical_doe_iterative():
    quantity_a = [
        ContinuousInput(key=f"quantity_a_{i}", bounds=(20, 100)) for i in range(2)
    ]
    all_inputs = [
        ContinuousInput(key="independent", bounds=(3, 10)),
    ]
    all_inputs.extend(quantity_a)

    all_constraints = [
        NChooseKConstraint(
            features=[var.key for var in quantity_a],
            min_count=1,
            max_count=1,
            none_also_valid=False,
        ),
    ]

    n_experiments = 5
    domain = Domain(
        inputs=all_inputs,
        outputs=[ContinuousOutput(key="y")],
        constraints=all_constraints,
    )

    data_model = data_models.DoEStrategy(
        domain=domain,
        formula="linear",
        optimization_strategy="iterative",
    )
    strategy = DoEStrategy(data_model=data_model)
    candidates = strategy.ask(
        candidate_count=n_experiments, raise_validation_error=False
    )

    assert candidates.shape == (5, 3)


if __name__ == "__main__":
    test_categorical_doe_iterative()



import gpytorch

import bofire.data_models.priors.api as data_models


def map_NormalPrior(data_model: data_models.NormalPrior) -> gpytorch.priors.NormalPrior:
    return gpytorch.priors.NormalPrior(loc=data_model.loc, scale=data_model.scale)


def map_GammaPrior(data_model: data_models.GammaPrior) -> gpytorch.priors.GammaPrior:
    return gpytorch.priors.GammaPrior(
        concentration=data_model.concentration, rate=data_model.rate
    )


PRIOR_MAP = {
    data_models.NormalPrior: map_NormalPrior,
    data_models.GammaPrior: map_GammaPrior,
}


def map(
    data_model: data_models.AnyPrior,
) -> gpytorch.priors.Prior:
    return PRIOR_MAP[data_model.__class__](data_model)



from typing import Literal

from pydantic import Field, validator

from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.kernels.api import (
    AnyCategoricalKernal,
    AnyContinuousKernel,
    HammondDistanceKernel,
    MaternKernel,
)
from bofire.data_models.surrogates.botorch import BotorchSurrogate
from bofire.data_models.surrogates.single_task_gp import ScalerEnum
from bofire.data_models.surrogates.trainable import TrainableSurrogate


class MixedSingleTaskGPSurrogate(BotorchSurrogate, TrainableSurrogate):
    type: Literal["MixedSingleTaskGPSurrogate"] = "MixedSingleTaskGPSurrogate"
    continuous_kernel: AnyContinuousKernel = Field(
        default_factory=lambda: MaternKernel(ard=True, nu=2.5)
    )
    categorical_kernel: AnyCategoricalKernal = Field(
        default_factory=lambda: HammondDistanceKernel(ard=True)
    )
    scaler: ScalerEnum = ScalerEnum.NORMALIZE

    @validator("input_preprocessing_specs")
    def validate_categoricals(cls, v, values):
        """Checks that at least one one-hot encoded categorical feauture is present."""
        if CategoricalEncodingEnum.ONE_HOT not in v.values():
            raise ValueError(
                "MixedSingleTaskGPSurrogate can only be used if at least one one-hot encoded categorical feature is present."
            )
        return v



from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from torch import Tensor

from bofire.data_models.api import AnyObjective, Domain, Outputs
from bofire.data_models.constraints.api import (
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
)
from bofire.data_models.features.api import ContinuousInput, Input
from bofire.data_models.objectives.api import (
    CloseToTargetObjective,
    ConstrainedObjective,
    MaximizeObjective,
    MaximizeSigmoidObjective,
    MinimizeObjective,
    MinimizeSigmoidObjective,
    TargetObjective,
)
from bofire.strategies.strategy import Strategy

tkwargs = {
    "dtype": torch.double,
    "device": "cpu",
}


def get_linear_constraints(
    domain: Domain,
    constraint: Union[LinearEqualityConstraint, LinearInequalityConstraint],
    unit_scaled: bool = False,
) -> List[Tuple[Tensor, Tensor, float]]:
    """Converts linear constraints to the form required by BoTorch.

    Args:
        domain (Domain): Optimization problem definition.
        constraint (Union[LinearEqualityConstraint, LinearInequalityConstraint]): Type of constraint that should be converted.
        unit_scaled (bool, optional): If True, transforms constraints by assuming that the bound for the continuous features are [0,1]. Defaults to False.

    Returns:
        List[Tuple[Tensor, Tensor, float]]: List of tuples, each tuple consists of a tensor with the feature indices, coefficients and a float for the rhs.
    """
    constraints = []
    for c in domain.constraints.get(constraint):
        indices = []
        coefficients = []
        lower = []
        upper = []
        rhs = 0.0
        for i, featkey in enumerate(c.features):  # type: ignore
            idx = domain.get_feature_keys(Input).index(featkey)
            feat = domain.get_feature(featkey)
            if feat.is_fixed():  # type: ignore
                rhs -= feat.fixed_value()[0] * c.coefficients[i]  # type: ignore
            else:
                lower.append(feat.lower_bound)  # type: ignore
                upper.append(feat.upper_bound)  # type: ignore
                indices.append(idx)
                coefficients.append(
                    c.coefficients[i]  # type: ignore
                )  # if unit_scaled == False else c_scaled.coefficients[i])
        if unit_scaled:
            lower = np.array(lower)
            upper = np.array(upper)
            s = upper - lower
            scaled_coefficients = s * np.array(coefficients)
            constraints.append(
                (
                    torch.tensor(indices),
                    -torch.tensor(scaled_coefficients).to(**tkwargs),
                    -(rhs + c.rhs - np.sum(np.array(coefficients) * lower)),  # type: ignore
                )
            )
        else:
            constraints.append(
                (
                    torch.tensor(indices),
                    -torch.tensor(coefficients).to(**tkwargs),
                    -(rhs + c.rhs),  # type: ignore
                )
            )
    return constraints


def get_nchoosek_constraints(domain: Domain) -> List[Callable[[Tensor], float]]:
    """Transforms NChooseK constraints into a list of non-linear inequality constraint callables
    that can be parsed by pydantic. For this purpose the NChooseK constraint is continuously
    relaxed by countig the number of zeros in a candidate by a sum of narrow gaussians centered
    at zero.

    Args:
        domain (Domain): Optimization problem definition.

    Returns:
        List[Callable[[Tensor], float]]: List of callables that can be used
            as nonlinear equality constraints in botorch.
    """

    def narrow_gaussian(x, ell=1e-3):
        return torch.exp(-0.5 * (x / ell) ** 2)

    def max_constraint(indices: Tensor, num_features: int, max_count: int):
        return lambda x: narrow_gaussian(x=x[..., indices]).sum(dim=-1) - (
            num_features - max_count
        )

    def min_constraint(indices: Tensor, num_features: int, min_count: int):
        return lambda x: -narrow_gaussian(x=x[..., indices]).sum(dim=-1) + (
            num_features - min_count
        )

    constraints = []
    # ignore none also valid for the start
    for c in domain.constraints.get(NChooseKConstraint):
        assert isinstance(c, NChooseKConstraint)
        indices = torch.tensor(
            [domain.get_feature_keys(ContinuousInput).index(key) for key in c.features],
            dtype=torch.int64,
        )
        if c.max_count != len(c.features):
            constraints.append(
                max_constraint(
                    indices=indices, num_features=len(c.features), max_count=c.max_count
                )
            )
        if c.min_count > 0:
            constraints.append(
                min_constraint(
                    indices=indices, num_features=len(c.features), min_count=c.min_count
                )
            )
    return constraints


def constrained_objective2botorch(
    idx: int,
    objective: ConstrainedObjective,
) -> Tuple[List[Callable[[Tensor], Tensor]], List[float]]:
    """Create a callable that can be used by `botorch.utils.objective.apply_constraints`
    to setup ouput constrained optimizations.

    Args:
        idx (int): Index of the constraint objective in the list of outputs.
        objective (BotorchConstrainedObjective): The objective that should be transformed.

    Returns:
        Tuple[List[Callable[[Tensor], Tensor]], List[float]]: List of callables that can be used by botorch for setting up the constrained objective, and
            list of the corresponding botorch eta values.
    """
    assert isinstance(
        objective, ConstrainedObjective
    ), "Objective is not a `ConstrainedObjective`."
    if isinstance(objective, MaximizeSigmoidObjective):
        return [lambda Z: (Z[..., idx] - objective.tp) * -1.0], [
            1.0 / objective.steepness
        ]
    elif isinstance(objective, MinimizeSigmoidObjective):
        return [lambda Z: (Z[..., idx] - objective.tp)], [1.0 / objective.steepness]
    elif isinstance(objective, TargetObjective):
        return [
            lambda Z: (Z[..., idx] - (objective.target_value - objective.tolerance))
            * -1.0,
            lambda Z: (Z[..., idx] - (objective.target_value + objective.tolerance)),
        ], [1.0 / objective.steepness, 1.0 / objective.steepness]
    else:
        raise ValueError(f"Objective {objective.__class__.__name__} not known.")


def get_output_constraints(
    outputs: Outputs,
) -> Tuple[List[Callable[[Tensor], Tensor]], List[float]]:
    """Method to translate output constraint objectives into a list of
    callables and list of etas for use in botorch.

    Args:
        outputs (Outputs): Output feature object that should
            be processed.

    Returns:
        Tuple[List[Callable[[Tensor], Tensor]], List[float]]: List of constraint callables,
            list of associated etas.
    """
    constraints = []
    etas = []
    for idx, feat in enumerate(outputs.get()):
        if isinstance(feat.objective, ConstrainedObjective):  # type: ignore
            iconstraints, ietas = constrained_objective2botorch(
                idx, objective=feat.objective  # type: ignore
            )
            constraints += iconstraints
            etas += ietas
    return constraints, etas


def get_objective_callable(
    idx: int, objective: AnyObjective
) -> Callable[[Tensor, Optional[Tensor]], Tensor]:  # type: ignore
    if isinstance(objective, MaximizeObjective):
        return lambda y, X=None: (
            (y[..., idx] - objective.lower_bound)
            / (objective.upper_bound - objective.lower_bound)
        )
    if isinstance(objective, MinimizeObjective):
        return lambda y, X=None: -1.0 * (
            (y[..., idx] - objective.lower_bound)
            / (objective.upper_bound - objective.lower_bound)
        )
    if isinstance(objective, CloseToTargetObjective):
        return lambda y, X=None: -1.0 * (
            torch.abs(y[..., idx] - objective.target_value) ** objective.exponent
        )
    if isinstance(objective, MinimizeSigmoidObjective):
        return lambda y, X=None: (
            (
                1.0
                - 1.0
                / (
                    1.0
                    + torch.exp(
                        -1.0 * objective.steepness * (y[..., idx] - objective.tp)
                    )
                )
            )
        )
    if isinstance(objective, MaximizeSigmoidObjective):
        return lambda y, X=None: (
            1.0
            / (
                1.0
                + torch.exp(-1.0 * objective.steepness * ((y[..., idx] - objective.tp)))
            )
        )
    if isinstance(objective, TargetObjective):
        return lambda y, X=None: (
            1.0
            / (
                1.0
                + torch.exp(
                    -1
                    * objective.steepness
                    * (y[..., idx] - (objective.target_value - objective.tolerance))
                )
            )
            * (
                1.0
                - 1.0
                / (
                    1.0
                    + torch.exp(
                        -1.0
                        * objective.steepness
                        * (y[..., idx] - (objective.target_value + objective.tolerance))
                    )
                )
            )
        )
    else:
        raise NotImplementedError(
            f"Objective {objective.__class__.__name__} not implemented."
        )


def get_custom_botorch_objective(
    outputs: Outputs,
    f: Callable[
        [
            Tensor,
            List[Callable[[Tensor, Optional[Tensor]], Tensor]],
            List[float],
            Tensor,
        ],
        Tensor,
    ],
    exclude_constraints: bool = True,
) -> Callable[[Tensor, Tensor], Tensor]:
    callables = [
        get_objective_callable(idx=i, objective=feat.objective)  # type: ignore
        for i, feat in enumerate(outputs.get())
        if feat.objective is not None  # type: ignore
        and (
            not isinstance(feat.objective, ConstrainedObjective)  # type: ignore
            if exclude_constraints
            else True
        )
    ]
    weights = [
        feat.objective.w  # type: ignore
        for i, feat in enumerate(outputs.get())
        if feat.objective is not None  # type: ignore
        and (
            not isinstance(feat.objective, ConstrainedObjective)  # type: ignore
            if exclude_constraints
            else True
        )
    ]

    def objective(samples: torch.Tensor, X: torch.Tensor) -> torch.Tensor:
        return f(samples, callables, weights, X)

    return objective


def get_multiplicative_botorch_objective(
    outputs: Outputs,
) -> Callable[[Tensor, Tensor], Tensor]:
    callables = [
        get_objective_callable(idx=i, objective=feat.objective)  # type: ignore
        for i, feat in enumerate(outputs.get())
        if feat.objective is not None  # type: ignore
    ]
    weights = [
        feat.objective.w  # type: ignore
        for i, feat in enumerate(outputs.get())
        if feat.objective is not None  # type: ignore
    ]

    def objective(samples: torch.Tensor, X: torch.Tensor) -> torch.Tensor:
        val = 1.0
        for c, w in zip(callables, weights):
            val *= c(samples, None) ** w
        return val  # type: ignore

    return objective


def get_additive_botorch_objective(
    outputs: Outputs, exclude_constraints: bool = True
) -> Callable[[Tensor, Tensor], Tensor]:
    callables = [
        get_objective_callable(idx=i, objective=feat.objective)  # type: ignore
        for i, feat in enumerate(outputs.get())
        if feat.objective is not None  # type: ignore
        and (
            not isinstance(feat.objective, ConstrainedObjective)  # type: ignore
            if exclude_constraints
            else True
        )
    ]
    weights = [
        feat.objective.w  # type: ignore
        for i, feat in enumerate(outputs.get())
        if feat.objective is not None  # type: ignore
        and (
            not isinstance(feat.objective, ConstrainedObjective)  # type: ignore
            if exclude_constraints
            else True
        )
    ]

    def objective(samples: Tensor, X: Tensor) -> Tensor:
        val = 0.0
        for c, w in zip(callables, weights):
            val += c(samples, None) * w
        return val  # type: ignore

    return objective


def get_multiobjective_objective(
    outputs: Outputs,
) -> Callable[[Tensor, Optional[Tensor]], Tensor]:
    """Returns

    Args:
        outputs (Outputs): _description_

    Returns:
        Callable[[Tensor], Tensor]: _description_
    """
    callables = [
        get_objective_callable(idx=i, objective=feat.objective)  # type: ignore
        for i, feat in enumerate(outputs.get())
        if feat.objective is not None  # type: ignore
        and isinstance(
            feat.objective,  # type: ignore
            (MaximizeObjective, MinimizeObjective, CloseToTargetObjective),
        )
    ]

    def objective(samples: Tensor, X: Optional[Tensor] = None) -> Tensor:
        return torch.stack([c(samples, None) for c in callables], dim=-1)

    return objective


def get_initial_conditions_generator(
    strategy: Strategy,
    transform_specs: Dict,
    ask_options: Optional[Dict] = None,
    sequential: bool = True,
) -> Callable[[int, int, int], Tensor]:
    """Takes a strategy object and returns a callable which uses this
    strategy to return a generator callable which can be used in botorch`s
    `gen_batch_initial_conditions` to generate samples.

    Args:
        strategy (Strategy): Strategy that should be used to generate samples.
        transform_specs (Dict): Dictionary indicating how the samples should be
            transformed.
        ask_options (Dict, optional): Dictionary of keyword arguments that are
            passed to the `ask` method of the strategy. Defaults to {}.
        sequential (bool, optional): If True, samples for every q-batch are
            generate indepenent from each other. If False, the `n x q` samples
            are generated at once.

    Returns:
        Callable[[int, int, int], Tensor]: Callable that can be passed to
            `batch_initial_conditions`.
    """
    if ask_options is None:
        ask_options = {}

    def generator(n: int, q: int, seed: int) -> Tensor:
        if sequential:
            initial_conditions = []
            for _ in range(n):
                candidates = strategy.ask(q, **ask_options)
                # transform it
                transformed_candidates = strategy.domain.inputs.transform(
                    candidates, transform_specs
                )
                # transform to tensor
                initial_conditions.append(
                    torch.from_numpy(transformed_candidates.values).to(**tkwargs)
                )
            return torch.stack(initial_conditions, dim=0)
        else:
            candidates = strategy.ask(n * q, **ask_options)
            # transform it
            transformed_candidates = strategy.domain.inputs.transform(
                candidates, transform_specs
            )
            return (
                torch.from_numpy(transformed_candidates.values)
                .to(**tkwargs)
                .reshape(n, q, transformed_candidates.shape[1])
            )

    return generator



import json
import math
from typing import Optional

import numpy as np
import pandas as pd
import torch
from pydantic import validator
from pydantic.types import PositiveInt
from scipy.integrate import solve_ivp
from scipy.special import gamma

import bofire.surrogates.api as surrogates
from bofire.benchmarks.benchmark import Benchmark
from bofire.benchmarks.data.aniline_cn_crosscoupling import (
    EXPERIMENTS as ANNILINE_CN_CROSSCOUPLING_EXPERIMENTS,
)
from bofire.data_models.domain.api import Domain, Inputs, Outputs
from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    ContinuousInput,
    ContinuousOutput,
    Input,
)
from bofire.data_models.objectives.api import (
    MaximizeObjective,
    MaximizeSigmoidObjective,
    MinimizeObjective,
)
from bofire.data_models.surrogates.api import SingleTaskGPSurrogate


class DTLZ2(Benchmark):
    """Multiobjective bechmark function for testing optimization algorithms.
    Info about the function: https://pymoo.org/problems/many/dtlz.html
    """

    def __init__(self, dim: PositiveInt, num_objectives: PositiveInt = 2, **kwargs):
        """Initiallizes object of Type DTLZ2 which is a benchmark function.

        Args:
            dim (PositiveInt): Dimension of input vector
            num_objectives (PositiveInt, optional): Dimension of output vector. Defaults to 2.
        """
        super().__init__(**kwargs)
        self.num_objectives = num_objectives
        self.dim = dim

        inputs = []
        for i in range(self.dim):
            inputs.append(ContinuousInput(key="x_%i" % (i), bounds=(0, 1)))
        outputs = []
        self.k = self.dim - self.num_objectives + 1
        for i in range(self.num_objectives):
            outputs.append(
                ContinuousOutput(key=f"f_{i}", objective=MinimizeObjective(w=1.0))
            )
        domain = Domain(
            inputs=Inputs(features=inputs),
            outputs=Outputs(features=outputs),
        )
        self.ref_point = {
            feat: 1.1 for feat in domain.get_feature_keys(ContinuousOutput)
        }
        self._domain = domain

    @validator("dim")
    def validate_dim(cls, dim, values):
        num_objectives = values["num_objectives"]
        if dim <= values["num_objectives"]:
            raise ValueError(
                f"dim must be > num_objectives, but got {dim} and {num_objectives}."
            )
        return dim

    @property
    def best_possible_hypervolume(self) -> float:
        # hypercube - volume of hypersphere in R^d such that all coordinates are
        # positive
        hypercube_vol = self.ref_point[0] ** self.num_objectives  # type: ignore
        pos_hypersphere_vol = (
            math.pi ** (self.num_objectives / 2)
            / gamma(self.num_objectives / 2 + 1)
            / 2**self.num_objectives
        )
        return hypercube_vol - pos_hypersphere_vol

    def _f(self, candidates: pd.DataFrame) -> pd.DataFrame:
        """Function evaluation of DTLZ2.

        Args:
            candidates (pd.DataFrame): Input vector for x-values. Columns go from x0 to xdim.

        Returns:
            pd.DataFrame: Function values in output vector. Columns are f0 and f1.
        """
        X = candidates[self.domain.get_feature_keys(Input)].values  # type: ignore
        X_m = X[..., -self.k :]  # type: ignore
        g_X = ((X_m - 0.5) ** 2).sum(axis=-1)  # type: ignore
        g_X_plus1 = 1 + g_X
        fs = []
        pi_over_2 = math.pi / 2
        for i in range(self.num_objectives):
            idx = self.num_objectives - 1 - i
            f_i = g_X_plus1.copy()
            f_i *= np.cos(X[..., :idx] * pi_over_2).prod(axis=-1)
            if i > 0:
                f_i *= np.sin(X[..., idx] * pi_over_2)
            fs.append(f_i)

        col_names = self.domain.outputs.get_keys_by_objective(includes=MinimizeObjective)  # type: ignore
        y_values = np.stack(fs, axis=-1)
        Y = pd.DataFrame(data=y_values, columns=col_names)
        Y[
            [
                "valid_%s" % feat
                for feat in self.domain.outputs.get_keys_by_objective(  # type: ignore
                    includes=MinimizeObjective
                )
            ]
        ] = 1
        return Y


class C2DTLZ2(DTLZ2):
    """Constrained DTLZ2 benchmark function. Taken from
    https://github.com/pytorch/botorch/blob/main/botorch/test_functions/multi_objective.py.
    """

    def __init__(self, dim: PositiveInt, num_objectives: PositiveInt = 2, **kwargs):
        super().__init__(dim, num_objectives, **kwargs)
        # add also the constraint
        self._domain.outputs.features.append(  # type: ignore
            ContinuousOutput(
                key="slack",
                objective=MaximizeSigmoidObjective(w=1.0, tp=0, steepness=1.0 / 1e-3),
            )
        )

    @property
    def best_possible_hypervolume(self) -> float:
        return 0.3996406303723544

    def _f(self, candidates: pd.DataFrame) -> pd.DataFrame:
        r = 0.2
        Y = super()._f(candidates=candidates)
        # here the constrained is calculated
        # TODO port it to numpy
        f_X = torch.from_numpy(
            Y[self.domain.outputs.get_keys_by_objective(MinimizeObjective)].values,
        )
        term1 = (f_X - 1).pow(2)
        mask = ~(torch.eye(f_X.shape[-1], device=f_X.device).bool())
        indices = torch.arange(f_X.shape[1], device=f_X.device).repeat(f_X.shape[1], 1)  # type: ignore
        indexer = indices[mask].view(f_X.shape[1], f_X.shape[-1] - 1)
        term2_inner = (
            f_X.unsqueeze(1)
            .expand(f_X.shape[0], f_X.shape[-1], f_X.shape[-1])
            .gather(dim=-1, index=indexer.repeat(f_X.shape[0], 1, 1))
        )
        term2 = (term2_inner.pow(2) - r**2).sum(dim=-1)
        min1 = (term1 + term2).min(dim=-1).values
        min2 = ((f_X - 1 / math.sqrt(f_X.shape[-1])).pow(2) - r**2).sum(dim=-1)
        slack = pd.Series(
            -torch.min(min1, min2).unsqueeze(-1).squeeze().numpy(), name="slack"
        )
        Y = pd.concat([Y, slack], axis=1)
        Y["valid_slack"] = 1
        return Y


class SnarBenchmark(Benchmark):
    """Nucleophilic aromatic substitution problem as a multiobjective test function for optimization algorithms.
    Solving of a differential equation system with varying intitial values.
    """

    def __init__(self, C_i: Optional[np.ndarray] = None, **kwargs):
        """Initializes multiobjective test function object of type SnarBenchmark.

        Args:
            C_i (Optional[np.ndarray]): Input concentrations. Defaults to [1, 1]
        """
        super().__init__(**kwargs)
        if C_i is None:
            C_i = np.array([1, 1])
        self.C_i = C_i

        # Decision variables
        # "residence time in minutes"
        inputs = [
            ContinuousInput(key="tau", bounds=(0.5, 2)),
            # "equivalents of pyrrolidine"
            ContinuousInput(key="equiv_pldn", bounds=(1, 5)),
            # "concentration of 2,4 dinitrofluorobenenze at reactor inlet (after mixing) in M"
            ContinuousInput(key="conc_dfnb", bounds=(0.1, 0.5)),
            # "Reactor temperature in degress celsius"
            ContinuousInput(key="temperature", bounds=(30, 120)),
        ]
        # Objectives
        # "space time yield (kg/m^3/h)"
        outputs = [
            ContinuousOutput(key="sty", objective=MaximizeObjective(w=1.0)),
            # "E-factor"
            ContinuousOutput(
                key="e_factor",
                objective=MinimizeObjective(w=1.0),
            ),
        ]
        self.ref_point = {"e_factor": 10.7, "sty": 2957.0}
        self._domain = Domain(
            inputs=Inputs(features=inputs),
            outputs=Outputs(features=outputs),
        )

    @property
    def best_possible_hypervolume(self):
        return 10000.0

    def _f(self, candidates: pd.DataFrame) -> pd.DataFrame:
        """Function evaluation. Returns output vector.

        Args:
            candidates (pd.DataFrame): Input vector. Columns: tau, equiv_pldn, conc_dfnb, temperature

        Returns:
            pd.DataFrame: Output vector. Columns: sty, e_factor
        """
        stys = []
        e_factors = []
        for _, candidate in candidates.iterrows():
            tau = float(candidate["tau"])
            equiv_pldn = float(candidate["equiv_pldn"])
            conc_dfnb = float(candidate["conc_dfnb"])
            T = float(candidate["temperature"])
            y, e_factor, res = self._integrate_equations(tau, equiv_pldn, conc_dfnb, T)
            stys.append(y)
            e_factors.append(e_factor)
            # candidates["sty"] = y
            # candidates["e_factor"] = e_factor

        # return only y values instead of appending them to input dataframe
        Y = pd.DataFrame({"sty": stys, "e_factor": e_factors})
        Y[
            [
                "valid_%s" % feat
                for feat in self.domain.outputs.get_keys_by_objective(  # type: ignore
                    excludes=None
                )
            ]
        ] = 1
        return Y

    def _integrate_equations(self, tau, equiv_pldn, conc_dfnb, temperature, **kwargs):
        # Initial Concentrations in mM
        self.C_i = np.zeros(5)
        self.C_i[0] = conc_dfnb
        self.C_i[1] = equiv_pldn * conc_dfnb

        # Flowrate and residence time
        V = 5  # mL
        q_tot = V / tau
        # C1_0 = kwargs.get("C1_0", 2.0)  # reservoir concentration of 1 is 1 M = 1 mM
        # C2_0 = kwargs.get("C2_0", 4.2)  # reservoir concentration of  2 is 2 M = 2 mM
        # q_1 = self.C_i[0] / C1_0 * q_tot  # flowrate of 1 (dfnb)
        # q_2 = self.C_i[1] / C2_0 * q_tot  # flowrate of 2 (pldn)
        # q_eth = q_tot - q_1 - q_2  # flowrate of ethanol

        # Integrate
        res = solve_ivp(self._integrand, [0, tau], self.C_i, args=(temperature,))
        C_final = res.y[:, -1]

        # # Add measurement noise
        # C_final += (
        #     C_final * self.rng.normal(scale=self.noise_level, size=len(C_final)) / 100
        # )
        # C_final[
        #     C_final < 0
        # ] = 0  # prevent negative values of concentration introduced by noise

        # Calculate STY and E-factor
        M = [159.09, 71.12, 210.21, 210.21, 261.33]  # molecular weights (g/mol)
        sty = 6e4 / 1000 * M[2] * C_final[2] * q_tot / V  # convert to kg m^-3 h^-1
        if sty < 1e-6:
            sty = 1e-6
        rho_eth = 0.789  # g/mL (should adjust to temp, but just using @ 25C)
        term_2 = 1e-3 * sum([M[i] * C_final[i] * q_tot for i in range(5) if i != 2])
        if np.isclose(C_final[2], 0.0):
            # Set to a large value if no product formed
            e_factor = 1e3
        else:
            e_factor = (q_tot * rho_eth + term_2) / (1e-3 * M[2] * C_final[2] * q_tot)
        if e_factor > 1e3:
            e_factor = 1e3

        return sty, e_factor, {}

    def _integrand(self, t, C, T):
        # Kinetic Constants
        R = 8.314 / 1000  # kJ/K/mol
        T_ref = 90 + 273.71  # Convert to deg K
        T = T + 273.71  # Convert to deg K
        # Need to convert from 10^-2 M^-1s^-1 to M^-1min^-1

        def k(k_ref, E_a, temp):
            return 0.6 * k_ref * np.exp(-E_a / R * (1 / temp - 1 / T_ref))

        k_a = k(57.9, 33.3, T)
        k_b = k(2.70, 35.3, T)
        k_c = k(0.865, 38.9, T)
        k_d = k(1.63, 44.8, T)

        # Reaction Rates
        r = np.zeros(5)
        for i in [0, 1]:  # Set to reactants when close
            C[i] = 0 if C[i] < 1e-6 * self.C_i[i] else C[i]  # type: ignore
        r[0] = -(k_a + k_b) * C[0] * C[1]
        r[1] = -(k_a + k_b) * C[0] * C[1] - k_c * C[1] * C[2] - k_d * C[1] * C[3]
        r[2] = k_a * C[0] * C[1] - k_c * C[1] * C[2]
        r[3] = k_a * C[0] * C[1] - k_d * C[1] * C[3]
        r[4] = k_c * C[1] * C[2] + k_d * C[1] * C[3]

        # Deltas
        dcdtau = r
        return dcdtau


class ZDT1(Benchmark):
    """ZDT1 function for testing optimization algorithms.
    Explanation of the function: https://datacrayon.com/posts/search-and-optimisation/practical-evolutionary-algorithms/synthetic-objective-functions-and-zdt1/
    """

    def __init__(self, n_inputs=30, **kwargs):
        """Initializes class of type ZDT1 which is a benchmark function for optimization problems.

        Args:
            n_inputs (int, optional): Number of inputs. Defaults to 30.
        """
        super().__init__(**kwargs)
        self.n_inputs = n_inputs
        inputs = [
            ContinuousInput(key=f"x{i+1}", bounds=(0, 1)) for i in range(n_inputs)
        ]
        inputs = Inputs(features=inputs)
        outputs = [
            ContinuousOutput(key=f"y{i+1}", objective=MinimizeObjective(w=1))
            for i in range(2)
        ]
        outputs = Outputs(features=outputs)
        self._domain = Domain(inputs=inputs, outputs=outputs)

    def _f(self, X: pd.DataFrame) -> pd.DataFrame:
        """Function evaluation.

        Args:
            X (pd.DataFrame): Input values

        Returns:
            pd.DataFrame: Function values. Columns are y1, y2, valid_y1 and valid_y2.
        """
        x = X[self._domain.inputs.get_keys()[1:]].to_numpy()
        g = 1 + 9 / (self.n_inputs - 1) * np.sum(x, axis=1)
        y1 = X["x1"].to_numpy()
        y2 = g * (1 - (y1 / g) ** 0.5)
        return pd.DataFrame(
            {"y1": y1, "y2": y2, "valid_y1": 1, "valid_y2": 1}, index=X.index
        )

    def get_optima(self, points=100) -> pd.DataFrame:
        """Pareto front of the output variables.

        Args:
            points (int, optional): Number of points of the pareto front. Defaults to 100.

        Returns:
            pd.DataFrame: 2D pareto front with x and y values.
        """
        x = np.linspace(0, 1, points)
        y = np.stack([x, 1 - np.sqrt(x)], axis=1)
        return pd.DataFrame(y, columns=self.domain.outputs.get_keys())


class CrossCoupling(Benchmark):
    """Cross Coupling adapted from Summit (https://github.com/sustainable-processes/summit)

    Virtual experiments representing the Aniline Cross-Coupling reaction
    similar to Baumgartner et al. (2019) <https://`doi.org/10.1021/acs.oprd.9b00236>.
    Experimental outcomes are based on a SingleTaskGP fitted on the experimental data published by Baumgartner et al.
    This is a five dimensional optimisation of temperature, residence time, base equivalents,
    catalyst and base.
    The categorical variables (catalyst and base) contain descriptors
    calculated using COSMO-RS. Specifically, the descriptors are the first two sigma moments.

    Args:
        Benchmark (Benchmark): Benchmark base class
    """

    def __init__(
        self,
        **kwargs,
    ):
        # "residence time in minutes"
        inputs = [
            CategoricalDescriptorInput(
                key="catalyst",
                categories=["tBuXPhos", "tBuBrettPhos", "AlPhos"],
                descriptors=[
                    "area_cat",
                    "M2_cat",
                ],  # , 'M3_cat', 'Macc3_cat', 'Mdon3_cat'] #,'mol_weight', 'sol']
                values=[
                    [
                        460.7543,
                        67.2057,
                    ],  # 30.8413, 2.3043, 0], #, 424.64, 421.25040226],
                    [
                        518.8408,
                        89.8738,
                    ],  # 39.4424, 2.5548, 0], #, 487.7, 781.11247064],
                    [
                        819.933,
                        129.0808,
                    ],  # 83.2017, 4.2959, 0], #, 815.06, 880.74916884],
                ],
            ),
            CategoricalDescriptorInput(
                key="base",
                categories=["TEA", "TMG", "BTMG", "DBU"],
                descriptors=[
                    "area",
                    "M2",
                ],  # , 'M3', 'Macc3', 'Mdon3', 'mol_weight', 'sol'
                values=[
                    [162.2992, 25.8165],  # 40.9469, 3.0278, 0], #101.19, 642.2973283],
                    [
                        165.5447,
                        81.4847,
                    ],  # 107.0287, 10.215, 0.0169], # 115.18, 534.01544123],
                    [
                        227.3523,
                        30.554,
                    ],  # 14.3676, 1.1196, 0.0127], # 171.28, 839.81215],
                    [192.4693, 59.8367],  # 82.0661, 7.42, 0], # 152.24, 1055.82799],
                ],
            ),
            # "base equivalents"
            ContinuousInput(key="base_eq", bounds=(1, 2.5)),
            # "Reactor temperature in degrees celsius"
            ContinuousInput(key="temperature", bounds=(30, 100)),
            # "residence time in seconds (s)"
            ContinuousInput(key="t_res", bounds=(60, 1800)),
        ]

        input_preprocessing_specs = {
            "catalyst": CategoricalEncodingEnum.DESCRIPTOR,
            "base": CategoricalEncodingEnum.DESCRIPTOR,
        }

        # Objectives: yield and cost
        outputs = [
            ContinuousOutput(
                key="yield",
                objective=MaximizeObjective(w=1.0, bounds=(0.0, 1.0)),
            ),
            ContinuousOutput(
                key="cost",
                objective=MinimizeObjective(w=1.0, bounds=(0.0, 1.0)),
            ),
        ]
        self.ref_point = {"yield": 0.0, "cost": 1.0}

        self._domain = Domain(
            inputs=Inputs(features=inputs),
            outputs=Outputs(features=outputs),
        )

        data = pd.DataFrame.from_dict(json.loads(ANNILINE_CN_CROSSCOUPLING_EXPERIMENTS))

        data = data.rename(columns={"base_equivalents": "base_eq", "yld": "yield"})
        data["valid_yield"] = 1

        data_model = SingleTaskGPSurrogate(
            inputs=Inputs(features=inputs),
            outputs=Outputs(features=[outputs[0]]),
            input_preprocessing_specs=input_preprocessing_specs,  # type: ignore
        )
        ground_truth_yield = surrogates.map(data_model)

        ground_truth_yield.fit(experiments=data)  # type: ignore
        self.ground_truth_yield = ground_truth_yield
        super().__init__(**kwargs)

    def _f(self, candidates: pd.DataFrame) -> pd.DataFrame:
        """Function evaluation. Returns output vector.

        Args:
            candidates (pd.DataFrame): Input vector. Columns: catalyst, base, base_eq, temperature, t_res

        Returns:
            pd.DataFrame: Output vector. Columns: yield, cost, valid_yield, valid_cost
        """

        costs = self._calculate_costs(candidates)
        yields = self.ground_truth_yield.predict(candidates)

        Y = pd.concat([yields["yield_pred"], pd.Series(costs, name="cost")], axis=1)
        Y.rename(columns={"yield_pred": "yield"}, inplace=True)

        Y[
            [
                "valid_%s" % feat
                for feat in self.domain.outputs.get_keys_by_objective(  # type: ignore
                    excludes=None
                )
            ]
        ] = 1
        return Y

    def _calculate_costs(self, conditions):
        """Function to calculate the overall costs of a recipe

        Args:
            conditions (pd.DataFrame): The suggested candidate experiments

        Returns:
            np.array: Vector with costs of suggested candidates
        """
        catalyst = conditions["catalyst"].values
        base = conditions["base"].values
        base_equiv = conditions["base_eq"].values

        # Calculate amounts
        droplet_vol = 40 * 1e-3  # mL
        mmol_triflate = 0.91 * droplet_vol
        mmol_anniline = 1.6 * mmol_triflate
        catalyst_equiv = {
            "tBuXPhos": 0.0095,
            "tBuBrettPhos": 0.0094,
            "AlPhos": 0.0094,
        }
        mmol_catalyst = [catalyst_equiv[c] * mmol_triflate for c in catalyst]
        mmol_base = base_equiv * mmol_triflate

        # Calculate costs
        cost_triflate = mmol_triflate * 5.91  # triflate is $5.91/mmol
        cost_anniline = mmol_anniline * 0.01  # anniline is $0.01/mmol
        cost_catalyst = np.array(
            [self._get_catalyst_cost(c, m) for c, m in zip(catalyst, mmol_catalyst)]
        )
        cost_base = np.array(
            [self._get_base_cost(b, m) for b, m in zip(base, mmol_base)]
        )
        tot_cost = cost_triflate + cost_anniline + cost_catalyst + cost_base
        if len(tot_cost) == 1:
            tot_cost = tot_cost[0]  # type: ignore
        return tot_cost

    def _get_catalyst_cost(self, catalyst, catalyst_mmol):
        """Function to calculate the catalyst costs

        Args:
            catalyst (str): Catalyst name
            catalyst_mmol (float): Amount of catalyst used

        Returns:
            float: Catalyst costs
        """
        catalyst_prices = {
            "tBuXPhos": 94.08,
            "tBuBrettPhos": 182.85,
            "AlPhos": 594.18,
        }
        return float(catalyst_prices[catalyst] * catalyst_mmol)

    def _get_base_cost(self, base, mmol_base):
        """Function to calculate the base costs

        Args:
            base (str): Base name
            mmol_base (float): Amount of base used

        Returns:
            float: Base costs
        """
        # prices in $/mmol
        base_prices = {
            "DBU": 0.03,
            "BTMG": 1.2,
            "TMG": 0.001,
            "TEA": 0.01,
        }
        return float(base_prices[base] * mmol_base)



import numpy as np
import pandas as pd
import pytest

from bofire.benchmarks.LookupTableBenchmark import LookupTableBenchmark
from bofire.data_models.domain.api import Domain, Inputs, Outputs
from bofire.data_models.features.api import (
    CategoricalInput,
    ContinuousOutput,
)
from bofire.data_models.objectives.api import MaximizeObjective


@pytest.mark.parametrize(
    "cls_benchmark",
    [
        (LookupTableBenchmark),
    ],
)
def test_LookupTable_benchmark(cls_benchmark: LookupTableBenchmark):
    """Tests the initializer of Aspen_benchmark and whether the filename, domain and paths are set up correctly.

    Args:
        cls_benchmark (Aspen_benchmark): Aspen_benchmark class
        return_complete (bool): _description_
        kwargs (dict): Arguments to the initializer of Aspen_benchmark. {"filename": , "domain": , "paths": }
    """

    look_up = pd.DataFrame(columns=["x1", "x2", "y"])
    look_up["x1"] = np.random.choice(["a", "b", "c", "d"], 10)
    look_up["x2"] = np.random.choice(["e", "b", "f", "d"], 10)
    look_up["y"] = np.random.rand(10)
    look_up = look_up.drop_duplicates(subset=["x1", "x2"])
    input_feature1 = CategoricalInput(
        key="x1",
        categories=list(set(look_up["x1"].to_list())),
    )
    input_feature2 = CategoricalInput(
        key="x2",
        categories=list(set(look_up["x2"].to_list())),
    )
    objective = MaximizeObjective(
        w=1.0,
    )
    inputs = Inputs(features=[input_feature1, input_feature2])
    output_feature = ContinuousOutput(key="y", objective=objective)
    outputs = Outputs(features=[output_feature])
    domain = Domain(inputs=inputs, outputs=outputs)
    benchmark_function = cls_benchmark(domain=domain, lookup_table=look_up)
    # Test, if domain gets set up correctly
    assert benchmark_function.domain == domain
    # Test, if lookup table gets set up correctly
    pd.testing.assert_frame_equal(benchmark_function.lookup_table, look_up)
    # Test, if domain raises error for wrong key
    output_feature = ContinuousOutput(key="y1", objective=objective)
    outputs = Outputs(features=[output_feature])
    domain = Domain(inputs=inputs, outputs=outputs)
    with pytest.raises(ValueError):
        benchmark_function = cls_benchmark(domain=domain, lookup_table=look_up)
    # Test, if the output function works
    inputs = Inputs(features=[input_feature1, input_feature2])
    output_feature = ContinuousOutput(key="y", objective=objective)
    outputs = Outputs(features=[output_feature])
    domain = Domain(inputs=inputs, outputs=outputs)
    benchmark_function = cls_benchmark(domain=domain, lookup_table=look_up)
    sampled = look_up.loc[:2, domain.inputs.get_keys()].copy()
    benchmark_function.f(sampled, return_complete=True)
    # Test, if output function raises error
    sampled.loc[1, "x2"] = "a"
    with pytest.raises(ValueError):
        benchmark_function.f(sampled, return_complete=True)



import numpy as np
import pandas as pd
import pytest

import bofire.surrogates.api as surrogates
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.features.api import ContinuousInput, ContinuousOutput
from bofire.data_models.kernels.api import RBFKernel, ScaleKernel
from bofire.data_models.surrogates.api import SingleTaskGPSurrogate
from bofire.surrogates.diagnostics import metrics
from bofire.surrogates.feature_importance import (
    combine_lengthscale_importances,
    combine_permutation_importances,
    lengthscale_importance,
    lengthscale_importance_hook,
    permutation_importance,
    permutation_importance_hook,
)


def get_model_and_data():
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(3)
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments = inputs.sample(n=20)
    experiments.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments["valid_y"] = 1
    model = SingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
    )
    model = surrogates.map(model)
    return model, experiments


def test_lengthscale_importance_invalid():
    model, experiments = get_model_and_data()
    surrogate_data = SingleTaskGPSurrogate(
        inputs=model.inputs, outputs=model.outputs, kernel=RBFKernel()
    )
    surrogate = surrogates.map(surrogate_data)
    surrogate.fit(experiments)
    with pytest.raises(ValueError, match="No lenghtscale based kernel found."):
        lengthscale_importance(surrogate=surrogate)
    surrogate_data = SingleTaskGPSurrogate(
        inputs=model.inputs,
        outputs=model.outputs,
        kernel=ScaleKernel(base_kernel=RBFKernel(ard=False)),
    )
    surrogate = surrogates.map(surrogate_data)
    surrogate.fit(experiments)
    with pytest.raises(ValueError, match="Only one lengthscale found, use `ard=True`."):
        lengthscale_importance(surrogate=surrogate)


def test_lengthscale_importance():
    surrogate, experiments = get_model_and_data()
    surrogate.fit(experiments)
    importance = lengthscale_importance(surrogate=surrogate)
    assert isinstance(importance, pd.Series)
    assert list(importance.index) == surrogate.inputs.get_keys()
    importance = lengthscale_importance_hook(surrogate=surrogate)
    assert isinstance(importance, pd.Series)
    assert list(importance.index) == surrogate.inputs.get_keys()


def test_combine_lengthscale_importances():
    importances = [
        pd.Series(index=["x_1", "x_2", "x_3"], data=np.random.uniform(size=3))
        for _ in range(5)
    ]
    combined = combine_lengthscale_importances(importances=importances)
    assert isinstance(combined, pd.DataFrame)
    assert combined.shape == (5, 3)
    assert list(combined.columns) == ["x_1", "x_2", "x_3"]


def test_permutation_importance_invalid():
    model, experiments = get_model_and_data()
    X = experiments[model.inputs.get_keys()]
    y = experiments[["y"]]
    model.fit(experiments=experiments)
    with pytest.raises(AssertionError):
        permutation_importance(surrogate=model, X=X, y=y, n_repeats=1)
    with pytest.raises(AssertionError):
        permutation_importance(surrogate=model, X=X, y=y, n_repeats=2, seed=-1)


def test_permutation_importance():
    model, experiments = get_model_and_data()
    X = experiments[model.inputs.get_keys()]
    y = experiments[["y"]]
    model.fit(experiments=experiments)
    results = permutation_importance(surrogate=model, X=X, y=y, n_repeats=5)
    assert isinstance(results, dict)
    assert len(results) == len(metrics)
    for m in metrics.keys():
        assert m.name in results.keys()
        assert isinstance(results[m.name], pd.DataFrame)
        assert list(results[m.name].columns) == model.inputs.get_keys()
        assert list(results[m.name].index) == ["mean", "std"]


def test_permutation_importance_nan():
    model, experiments = get_model_and_data()
    X = experiments[model.inputs.get_keys()][:1]
    y = experiments[["y"]][:1]
    model.fit(experiments=experiments)
    results = permutation_importance(surrogate=model, X=X, y=y, n_repeats=5)
    assert isinstance(results, dict)
    assert len(results) == len(metrics)
    for m in metrics.keys():
        assert m.name in results.keys()
        assert isinstance(results[m.name], pd.DataFrame)
        assert list(results[m.name].columns) == model.inputs.get_keys()
        assert list(results[m.name].index) == ["mean", "std"]
        assert len(results[m.name].dropna()) == 0


@pytest.mark.parametrize("use_test", [True, False])
def test_permutation_importance_hook(use_test):
    model, experiments = get_model_and_data()
    X = experiments[model.inputs.get_keys()]
    y = experiments[["y"]]
    model.fit(experiments=experiments)
    results = permutation_importance_hook(
        surrogate=model, X_train=X, y_train=y, X_test=X, y_test=y, use_test=use_test
    )
    assert isinstance(results, dict)
    assert len(results) == len(metrics)
    for m in metrics.keys():
        assert m.name in results.keys()
        assert isinstance(results[m.name], pd.DataFrame)
        assert list(results[m.name].columns) == model.inputs.get_keys()
        assert list(results[m.name].index) == ["mean", "std"]


@pytest.mark.parametrize("n_folds", [5, 3])
def test_combine_permutation_importances(n_folds):
    model, experiments = get_model_and_data()
    _, _, pi = model.cross_validate(
        experiments,
        folds=n_folds,
        hooks={"pemutation_importance": permutation_importance_hook},
    )
    for m in metrics.keys():
        importance = combine_permutation_importances(
            importances=pi["pemutation_importance"], metric=m
        )
        assert list(importance.columns) == model.inputs.get_keys()
        assert len(importance) == n_folds



from bofire.plot.feature_importance import (  # noqa: F401
    plot_feature_importance_by_feature_plotly,
)
from bofire.plot.objective import plot_objective_plotly  # noqa: F401
from bofire.plot.prior import plot_prior_pdf_plotly  # noqa: F401



# TODO: remove this and use tests.bofire.data_models.specs instead

import uuid
from typing import List

from bofire.data_models.constraints.api import (
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
    NonlinearEqualityConstraint,
    NonlinearInequalityConstraint,
)
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    ContinuousDescriptorInput,
    ContinuousInput,
    ContinuousOutput,
    DiscreteInput,
)
from bofire.data_models.objectives.api import MinimizeObjective


def get_invalids(valid: dict) -> List[dict]:
    return [
        {k: v for k, v in valid.items() if k != k_}
        for k_ in valid.keys()
        if k_ != "type"
    ]


INVALID_SPECS = [
    [1, 2, 3],
    {"a": "qwe"},
]


objective = MinimizeObjective(w=1)

VALID_CONTINUOUS_INPUT_FEATURE_SPEC = {
    "type": "ContinuousInput",
    "key": str(uuid.uuid4()),
    "bounds": (3, 5.3),
}

VALID_FIXED_CONTINUOUS_INPUT_FEATURE_SPEC = {
    "type": "ContinuousInput",
    "key": str(uuid.uuid4()),
    "bounds": (3, 3),
}

VALID_DISCRETE_INPUT_FEATURE_SPEC = {
    "type": "DiscreteInput",
    "key": str(uuid.uuid4()),
    "values": [1.0, 2.0],
}

VALID_FIXED_DISCRETE_INPUT_FEATURE_SPEC = {
    "type": "DiscreteInput",
    "key": str(uuid.uuid4()),
    "values": [2.0],
}

VALID_CONTINUOUS_DESCRIPTOR_INPUT_FEATURE_SPEC = {
    "type": "ContinuousDescriptorInput",
    "key": str(uuid.uuid4()),
    "bounds": (3, 5.3),
    "descriptors": ["d1", "d2"],
    "values": [1.0, 2.0],
}

VALID_CATEGORICAL_INPUT_FEATURE_SPEC = {
    "type": "CategoricalInput",
    "key": str(uuid.uuid4()),
    "categories": ["c1", "c2", "c3"],
    # "allowed": [True, True, False],
}

VALID_CATEGORICAL_DESCRIPTOR_INPUT_FEATURE_SPEC = {
    "type": "CategoricalDescriptorInput",
    "key": str(uuid.uuid4()),
    "categories": ["c1", "c2", "c3"],
    # "allowed": [True, True, False],
    "descriptors": ["d1", "d2"],
    "values": [
        [1, 2],
        [3, 7],
        [5, 1],
    ],
}


VALID_ALLOWED_CATEGORICAL_DESCRIPTOR_INPUT_FEATURE_SPEC = {
    "type": "CategoricalDescriptorInput",
    "key": str(uuid.uuid4()),
    "categories": ["c1", "c2", "c3"],
    "allowed": [False, True, True],
    "descriptors": ["d1", "d2"],
    "values": [
        [1, 2],
        [3, 7],
        [3, 1],
    ],
}

VALID_FIXED_CATEGORICAL_INPUT_FEATURE_SPEC = {
    "type": "CategoricalInput",
    "key": str(uuid.uuid4()),
    "categories": ["c1", "c2", "c3"],
    "allowed": [True, False, False],
}

VALID_FIXED_CATEGORICAL_DESCRIPTOR_INPUT_FEATURE_SPEC = {
    "type": "CategoricalDescriptorInput",
    "key": str(uuid.uuid4()),
    "categories": ["c1", "c2", "c3"],
    "allowed": [True, False, False],
    "descriptors": ["d1", "d2"],
    "values": [
        [1, 2],
        [3, 7],
        [5, 1],
    ],
}

VALID_CONTINUOUS_OUTPUT_FEATURE_SPEC = {
    "type": "ContinuousOutput",
    "key": str(uuid.uuid4()),
}

FEATURE_SPECS = {
    ContinuousInput: {
        "valids": [
            VALID_CONTINUOUS_INPUT_FEATURE_SPEC,
            VALID_FIXED_CONTINUOUS_INPUT_FEATURE_SPEC,
        ],
        "invalids": [
            *get_invalids(VALID_CONTINUOUS_INPUT_FEATURE_SPEC),
        ],
    },
    DiscreteInput: {
        "valids": [
            VALID_DISCRETE_INPUT_FEATURE_SPEC,
            VALID_FIXED_DISCRETE_INPUT_FEATURE_SPEC,
        ],
        "invalids": [
            *get_invalids(VALID_DISCRETE_INPUT_FEATURE_SPEC),
            *[
                {
                    **VALID_DISCRETE_INPUT_FEATURE_SPEC,
                    "values": values,
                }
                for values in [[], [1.0, 1.0], [1.0, "a"]]
            ],
        ],
    },
    ContinuousDescriptorInput: {
        "valids": [VALID_CONTINUOUS_DESCRIPTOR_INPUT_FEATURE_SPEC],
        "invalids": [
            *get_invalids(VALID_CONTINUOUS_DESCRIPTOR_INPUT_FEATURE_SPEC),
            *[
                {
                    **VALID_CONTINUOUS_DESCRIPTOR_INPUT_FEATURE_SPEC,
                    "descriptors": descriptors,
                    "values": values,
                }
                for descriptors, values in [
                    ([], []),
                    (["a", "b"], [1]),
                    (["a", "b"], [1, 2, 3]),
                ]
            ],
        ],
    },
    CategoricalInput: {
        "valids": [
            VALID_CATEGORICAL_INPUT_FEATURE_SPEC,
            {
                **VALID_CATEGORICAL_INPUT_FEATURE_SPEC,
                "allowed": [True, False, True],
            },
            VALID_FIXED_CATEGORICAL_INPUT_FEATURE_SPEC,
        ],
        "invalids": [
            *get_invalids(VALID_CATEGORICAL_INPUT_FEATURE_SPEC),
            *[
                {
                    **VALID_CATEGORICAL_INPUT_FEATURE_SPEC,
                    "categories": categories,
                    "allowed": allowed,
                }
                for categories, allowed in [
                    ([], []),
                    (["1"], [False]),
                    (["1", "2"], [False, False]),
                    (["1", "1"], None),
                    (["1", "1", "2"], None),
                    (["1", "2"], [True]),
                    (["1", "2"], [True, False, True]),
                    (["1"], []),
                    (["1"], [True]),
                ]
            ],
        ],
    },
    CategoricalDescriptorInput: {
        "valids": [
            VALID_CATEGORICAL_DESCRIPTOR_INPUT_FEATURE_SPEC,
            {
                **VALID_CATEGORICAL_DESCRIPTOR_INPUT_FEATURE_SPEC,
                "allowed": [True, False, True],
            },
            VALID_FIXED_CATEGORICAL_DESCRIPTOR_INPUT_FEATURE_SPEC,
        ],
        "invalids": [
            *get_invalids(VALID_CATEGORICAL_DESCRIPTOR_INPUT_FEATURE_SPEC),
            *[
                {
                    **VALID_CATEGORICAL_INPUT_FEATURE_SPEC,
                    "categories": categories,
                    "descriptors": descriptors,
                    "values": values,
                }
                for categories, descriptors, values in [
                    (["c1", "c2"], ["d1", "d2", "d3"], []),
                    (["c1", "c2"], ["d1", "d2", "d3"], [[1, 2, 3]]),
                    (
                        ["c1", "c2"],
                        ["d1", "d2", "d3"],
                        [[1, 2, 3], [1, 2, 3], [1, 2, 3]],
                    ),
                    (["c1", "c2"], ["d1", "d2", "d3"], [[1, 2, 3], [1, 2]]),
                    (["c1", "c2"], ["d1", "d2", "d3"], [[1, 2, 3], [1, 2, 3, 4]]),
                    (["c1", "c2"], ["d1", "d2", "d3"], [[1, 2, 3], [1, 2, 3]]),
                ]
            ],
        ],
    },
    ContinuousOutput: {
        "valids": [
            VALID_CONTINUOUS_OUTPUT_FEATURE_SPEC,
            {
                **VALID_CONTINUOUS_OUTPUT_FEATURE_SPEC,
                "objective": objective,
            },
            {
                **VALID_CONTINUOUS_OUTPUT_FEATURE_SPEC,
                "objective": None,
            },
        ],
        "invalids": [*get_invalids(VALID_CONTINUOUS_OUTPUT_FEATURE_SPEC)],
    },
}


VALID_NCHOOSEKE_CONSTRAINT_SPEC = {
    "features": ["f1", "f2", "f3"],
    "min_count": 1,
    "max_count": 1,
    "none_also_valid": False,
}

VALID_LINEAR_EQUALITY_CONSTRAINT_SPEC = {
    "type": "LinearEqualityConstraint",
    "features": ["f1", "f2", "f3"],
    "coefficients": [1, 2, 3],
    "rhs": 1.5,
}

VALID_LINEAR_INEQUALITY_CONSTRAINT_SPEC = {
    "type": "LinearInequalityConstraint",
    "features": ["f1", "f2", "f3"],
    "coefficients": [1, 2, 3],
    "rhs": 1.5,
}

VALID_NONLINEAR_EQUALITY_CONSTRAINT_SPEC = {
    "type": "NonlinearEqualityConstraint",
    "features": ["f1", "f2"],
    "expression": "f1*f2",
}

VALID_NONLINEAR_INEQUALITY_CONSTRAINT_SPEC = {
    "type": "NonlinearInequalityConstraint",
    "features": ["f1", "f2"],
    "expression": "f1*f2",
}

INVALID_NONLINEAR_EQUALITY_CONSTRAINT_SPEC = {
    "type": "NonlinearEqualityConstraint",
    "expression": [5, 7, 8],
}

INVALID_NONLINEAR_INEQUALITY_CONSTRAINT_SPEC = {
    "type": "NonlinearInequalityConstraint",
    "expression": [5, 7, 8],
}

VALID_LINEAR_EQUALITY_CONSTRAINT_SPECS = [
    {
        **VALID_LINEAR_EQUALITY_CONSTRAINT_SPEC,
        "features": features,
        "coefficients": coefficients,
    }
    for features, coefficients in [
        (["f1", "f2"], [-0.4, 1.4]),
    ]
]

VALID_LINEAR_INEQUALITY_CONSTRAINT_SPECS = [
    {
        **VALID_LINEAR_INEQUALITY_CONSTRAINT_SPEC,
        "features": features,
        "coefficients": coefficients,
    }
    for features, coefficients in [
        (["f1", "f2"], [-0.4, 1.4]),
    ]
]

INVALID_LINEAR_EQUALITY_CONSTRAINT_SPECS = [
    {
        **VALID_LINEAR_EQUALITY_CONSTRAINT_SPEC,
        "features": features,
        "coefficients": coefficients,
    }
    for features, coefficients in [
        ([], []),
        ([], [1]),
        (["f1", "f2"], [-0.4]),
        (["f1", "f2"], [-0.4, 1.4, 4.3]),
        (["f1", "f1"], [1, 1]),
        (["f1", "f1", "f2"], [1, 1, 1]),
    ]
]

INVALID_LINEAR_INEQUALITY_CONSTRAINT_SPECS = [
    {
        **VALID_LINEAR_INEQUALITY_CONSTRAINT_SPEC,
        "features": features,
        "coefficients": coefficients,
    }
    for features, coefficients in [
        ([], []),
        ([], [1]),
        (["f1", "f2"], [-0.4]),
        (["f1", "f2"], [-0.4, 1.4, 4.3]),
        (["f1", "f1"], [1, 1]),
        (["f1", "f1", "f2"], [1, 1, 1]),
    ]
]


CONSTRAINT_SPECS = {
    NChooseKConstraint: {
        "valids": [VALID_NCHOOSEKE_CONSTRAINT_SPEC],
        "invalids": INVALID_SPECS + get_invalids(VALID_NCHOOSEKE_CONSTRAINT_SPEC),
    },
    LinearEqualityConstraint: {
        "valids": [VALID_LINEAR_EQUALITY_CONSTRAINT_SPEC]
        + VALID_LINEAR_EQUALITY_CONSTRAINT_SPECS,
        "invalids": INVALID_SPECS
        + get_invalids(VALID_LINEAR_EQUALITY_CONSTRAINT_SPEC)
        + INVALID_LINEAR_EQUALITY_CONSTRAINT_SPECS,
    },
    LinearInequalityConstraint: {
        "valids": [VALID_LINEAR_INEQUALITY_CONSTRAINT_SPEC]
        + VALID_LINEAR_INEQUALITY_CONSTRAINT_SPECS,
        "invalids": INVALID_SPECS
        + get_invalids(VALID_LINEAR_INEQUALITY_CONSTRAINT_SPEC)
        + INVALID_LINEAR_INEQUALITY_CONSTRAINT_SPECS,
    },
    NonlinearEqualityConstraint: {
        "valids": [VALID_NONLINEAR_EQUALITY_CONSTRAINT_SPEC],
        "invalids": [INVALID_NONLINEAR_EQUALITY_CONSTRAINT_SPEC],
    },
    NonlinearInequalityConstraint: {
        "valids": [VALID_NONLINEAR_INEQUALITY_CONSTRAINT_SPEC],
        "invalids": [INVALID_NONLINEAR_INEQUALITY_CONSTRAINT_SPEC],
    },
}



from enum import Enum


class SamplingMethodEnum(Enum):
    UNIFORM = "UNIFORM"
    SOBOL = "SOBOL"
    LHS = "LHS"


class CategoricalMethodEnum(Enum):
    """Enumeration class of supported methods how to handle categorical features
    Currently, exhaustive search and free relaxation are implemented.
    """

    EXHAUSTIVE = "EXHAUSTIVE"
    FREE = "FREE"
    # PR = "PR" available soon


class CategoricalEncodingEnum(Enum):
    """Enumeration class of implemented categorical encodings
    Currently, one-hot and ordinal encoding are implemented.
    """

    ONE_HOT = "ONE_HOT"
    ORDINAL = "ORDINAL"
    DUMMY = "DUMMY"
    DESCRIPTOR = "DESCRIPTOR"  # only possible for categorical with descriptors


class OutputFilteringEnum(Enum):
    ALL = "ALL"
    ANY = "ANY"


class RegressionMetricsEnum(Enum):
    """Enumeration class for regression metrics."""

    R2 = "R2"
    MAE = "MAE"
    MSD = "MSD"
    MAPE = "MAPE"
    PEARSON = "PEARSON"
    SPEARMAN = "SPEARMAN"
    FISHER = "FISHER"


class UQRegressionMetricsEnum(Enum):
    """Enumeration class for ucertainty regression metrics."""

    PEARSON_UQ = "PEARSON_UQ"
    SPEARMAN_UQ = "SPEARMAN_UQ"
    KENDALL_UQ = "KENDALL_UQ"
    MAXIMUMCALIBRATION = "MAXIMUMCALIBRATION"
    MISCALIBRATIONAREA = "MISCALIBRATIONAREA"
    ABSOLUTEMISCALIBRATIONAREA = "ABSOLUTEMISCALIBRATIONAREA"



from pytest import fixture

import tests.bofire.data_models.specs.api as specs


@fixture(params=specs.constraints.valids)
def constraint_spec(request) -> specs.Spec:
    return request.param


@fixture(params=specs.objectives.valids)
def objective_spec(request) -> specs.Spec:
    return request.param


@fixture(params=specs.features.valids)
def feature_spec(request) -> specs.Spec:
    return request.param


@fixture(params=specs.domain.valids)
def domain_spec(request) -> specs.Spec:
    return request.param


@fixture(params=specs.acquisition_functions.valids)
def acquisition_function_spec(request) -> specs.Spec:
    return request.param


try:
    # in case of the minimal installation these fixtures are not available
    @fixture(params=specs.strategies.valids)
    def strategy_spec(request) -> specs.Spec:
        return request.param

    @fixture(params=specs.surrogates.valids)
    def surrogate_spec(request) -> specs.Spec:
        return request.param

    @fixture(params=specs.priors.valids)
    def prior_spec(request) -> specs.Spec:
        return request.param

    @fixture(params=specs.kernels.valids)
    def kernel_spec(request) -> specs.Spec:
        return request.param

    @fixture(params=specs.conditions.valids)
    def condition_spec(request) -> specs.Spec:
        return request.param

    @fixture(params=specs.outlier_detection.valids)
    def outlier_detection_spec(request) -> specs.Spec:
        return request.param

    @fixture(params=specs.molfeatures.valids)
    def molfeatures_spec(request) -> specs.Spec:
        return request.param

except AttributeError:
    pass



import warnings
from typing import List, Optional

import numpy as np
import pandas as pd

try:
    from rdkit.Chem import AllChem, Descriptors, MolFromSmiles  # type: ignore

    # from sklearn.feature_extraction.text import CountVectorizer
except ImportError:
    warnings.warn(
        "rdkit not installed, BoFire's cheminformatics utilities cannot be used."
    )

try:
    from mordred import Calculator, descriptors
except ImportError:
    warnings.warn(
        "mordred not installed. Mordred molecular descriptors cannot be used."
    )

# This code is based on GAUCHE: https://github.com/leojklarner/gauche/blob/main/gauche/data_featuriser/featurisation.py


def smiles2mol(smiles: str):
    """Transforms a smiles string to an rdkit mol object.

    Args:
        smiles (str): Smiles string.

    Raises:
        ValueError: If string is not a valid smiles.

    Returns:
        rdkit.Mol: rdkit.mol object
    """
    mol = MolFromSmiles(smiles)
    if mol is None:
        raise ValueError(f"{smiles} is not a valid smiles string.")
    return mol


def smiles2fingerprints(
    smiles: List[str], bond_radius: int = 5, n_bits: int = 2048
) -> np.ndarray:
    """Transforms a list of smiles to an array of morgan fingerprints.

    Args:
        smiles (List[str]): List of smiles
        bond_radius (int, optional): Bond radius to use. Defaults to 5.
        n_bits (int, optional): Number of bits. Defaults to 2048.

    Returns:
        np.ndarray: Numpy array holding the fingerprints
    """
    rdkit_mols = [smiles2mol(m) for m in smiles]
    fps = [
        AllChem.GetMorganFingerprintAsBitVect(  # type: ignore
            mol, radius=bond_radius, nBits=n_bits
        )
        for mol in rdkit_mols
    ]

    return np.asarray(fps)


def smiles2fragments(
    smiles: List[str], fragments_list: Optional[List[str]] = None
) -> np.ndarray:
    """Transforms smiles to an array of fragments.

    Args:
        smiles (List[str]): List of smiles

    Returns:
        np.ndarray: Array holding the fragment information.
    """
    rdkit_fragment_list = [
        item for item in Descriptors.descList if item[0].startswith("fr_")
    ]
    if fragments_list is None:
        fragments = {d[0]: d[1] for d in rdkit_fragment_list}
    else:
        fragments = {d[0]: d[1] for d in rdkit_fragment_list if d[0] in fragments_list}

    frags = np.zeros((len(smiles), len(fragments)))
    for i, smi in enumerate(smiles):
        mol = smiles2mol(smi)
        features = [fragments[d](mol) for d in fragments]
        frags[i, :] = features

    return frags


# def smiles2bag_of_characters(smiles: List[str], max_ngram: int = 5) -> np.ndarray:
#     """Transforms list of smiles to bag of characters.
#
#     Args:
#         smiles (List[str]): List of smiles
#         max_ngram (int, optional): Maximal ngram value. Defaults to 5.
#
#     Returns:
#         np.ndarray: Array holding the bag of characters.
#     """
#     for smi in smiles:
#         smiles2mol(smi)
#     cv = CountVectorizer(ngram_range=(1, max_ngram), analyzer="char", lowercase=False)
#     return cv.fit_transform(smiles).toarray()


def smiles2mordred(smiles: List[str], descriptors_list: List[str]) -> np.ndarray:
    """Transforms list of smiles to mordred moelcular descriptors.

    Args:
        smiles (List[str]): List of smiles
        descriptors_list (List[str]): List of desired mordred descriptors

    Returns:
        np.ndarray: Array holding the mordred moelcular descriptors.
    """
    mols = [smiles2mol(smi) for smi in smiles]

    calc = Calculator(descriptors, ignore_3D=True)
    calc.descriptors = [d for d in calc.descriptors if str(d) in descriptors_list]

    descriptors_df = calc.pandas(mols)
    nan_list = [
        pd.to_numeric(descriptors_df[col], errors="coerce").isnull().values.any()
        for col in descriptors_df.columns
    ]
    if any(nan_list):
        raise ValueError(
            f"Found NaN values in descriptors {list(descriptors_df.columns[nan_list])}"
        )

    return descriptors_df.astype(float).values


def smiles2fragments_fingerprints(
    smiles: List[str],
    bond_radius: int = 5,
    n_bits: int = 2048,
    fragments_list: Optional[List[str]] = None,
) -> np.ndarray:
    fingerprints = smiles2fingerprints(smiles, bond_radius=bond_radius, n_bits=n_bits)
    fragments = smiles2fragments(smiles, fragments_list=fragments_list)

    return np.hstack((fingerprints, fragments))



from typing import Optional

import pandas as pd

import bofire.surrogates.api as surrogates
from bofire.benchmarks.benchmark import Benchmark
from bofire.data_models.domain.api import Domain
from bofire.data_models.surrogates.api import AnyTrainableSurrogate


class Hyperopt(Benchmark):
    def __init__(
        self,
        surrogate_data: AnyTrainableSurrogate,
        training_data: pd.DataFrame,
        folds: int,
        random_state: Optional[int] = None,
    ) -> None:
        super().__init__()
        if surrogate_data.hyperconfig is None:
            raise ValueError("No hyperoptimization configuration found.")
        self.surrogate_data = surrogate_data
        self.training_data = training_data
        self.folds = folds
        self.results = None
        self.random_state = random_state

    @property
    def domain(self) -> Domain:
        return self.surrogate_data.hyperconfig.domain  # type: ignore

    @property
    def target_metric(self):
        return self.surrogate_data.hyperconfig.target_metric  # type: ignore

    def _f(self, candidates: pd.DataFrame) -> pd.DataFrame:
        for i, candidate in candidates.iterrows():
            self.surrogate_data.update_hyperparameters(candidate)
            surrogate = surrogates.map(self.surrogate_data)
            _, cv_test, _ = surrogate.cross_validate(  # type: ignore
                self.training_data, folds=self.folds, random_state=self.random_state
            )
            if i == 0:
                results = cv_test.get_metrics(combine_folds=True)
            else:
                results = pd.concat(
                    [results, cv_test.get_metrics(combine_folds=True)],  # type: ignore
                    ignore_index=True,
                    axis=0,
                )
        results[f"valid_{self.target_metric.value}"] = 1  # type: ignore
        return results  # type: ignore



from typing import ClassVar, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import pandas as pd
from pydantic import Field, root_validator, validator
from typing_extensions import Annotated

from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.features.feature import (
    _CAT_SEP,
    Input,
    Output,
    TAllowedVals,
    TCategoryVals,
    TTransform,
)


class CategoricalInput(Input):
    """Base class for all categorical input features.

    Attributes:
        categories (List[str]): Names of the categories.
        allowed (List[bool]): List of bools indicating if a category is allowed within the optimization.
    """

    type: Literal["CategoricalInput"] = "CategoricalInput"
    order_id: ClassVar[int] = 5

    categories: TCategoryVals
    allowed: TAllowedVals = None

    @validator("categories")
    def validate_categories_unique(cls, categories):
        """validates that categories have unique names

        Args:
            categories (List[str]): List of category names

        Raises:
            ValueError: when categories have non-unique names

        Returns:
            List[str]: List of the categories
        """
        categories = list(categories)
        if len(categories) != len(set(categories)):
            raise ValueError("categories must be unique")
        return categories

    @root_validator(pre=False, skip_on_failure=True)
    def init_allowed(cls, values):
        """validates the list of allowed/not allowed categories

        Args:
            values (Dict): Dictionary with attributes

        Raises:
            ValueError: when the number of allowences does not fit to the number of categories
            ValueError: when no category is allowed

        Returns:
            Dict: Dictionary with attributes
        """
        if "categories" not in values or values["categories"] is None:
            return values
        if "allowed" not in values or values["allowed"] is None:
            values["allowed"] = [True for _ in range(len(values["categories"]))]
        if len(values["allowed"]) != len(values["categories"]):
            raise ValueError("allowed must have same length as categories")
        if sum(values["allowed"]) == 0:
            raise ValueError("no category is allowed")
        return values

    @staticmethod
    def valid_transform_types() -> List[CategoricalEncodingEnum]:
        return [
            CategoricalEncodingEnum.ONE_HOT,
            CategoricalEncodingEnum.DUMMY,
            CategoricalEncodingEnum.ORDINAL,
        ]

    def is_fixed(self) -> bool:
        """Returns True if there is only one allowed category.

        Returns:
            [bool]: True if there is only one allowed category
        """
        if self.allowed is None:
            return False
        return sum(self.allowed) == 1

    def fixed_value(
        self, transform_type: Optional[TTransform] = None
    ) -> Union[List[str], List[float], None]:
        """Returns the categories to which the feature is fixed, None if the feature is not fixed

        Returns:
            List[str]: List of categories or None
        """
        if self.is_fixed():
            val = self.get_allowed_categories()[0]
            if transform_type is None:
                return [val]
            elif transform_type == CategoricalEncodingEnum.ONE_HOT:
                return self.to_onehot_encoding(pd.Series([val])).values[0].tolist()
            elif transform_type == CategoricalEncodingEnum.DUMMY:
                return self.to_dummy_encoding(pd.Series([val])).values[0].tolist()
            elif transform_type == CategoricalEncodingEnum.ORDINAL:
                return self.to_ordinal_encoding(pd.Series([val])).tolist()
            else:
                raise ValueError(
                    f"Unkwon transform type {transform_type} for categorical input {self.key}"
                )
        else:
            return None

    def get_allowed_categories(self):
        """Returns the allowed categories.

        Returns:
            list of str: The allowed categories
        """
        if self.allowed is None:
            return []
        return [c for c, a in zip(self.categories, self.allowed) if a]

    def validate_experimental(
        self, values: pd.Series, strict: bool = False
    ) -> pd.Series:
        """Method to validate the experimental dataFrame

        Args:
            values (pd.Series): A dataFrame with experiments
            strict (bool, optional): Boolean to distinguish if the occurence of fixed features in the dataset should be considered or not. Defaults to False.

        Raises:
            ValueError: when an entry is not in the list of allowed categories
            ValueError: when there is no variation in a feature provided by the experimental data

        Returns:
            pd.Series: A dataFrame with experiments
        """
        values = values.map(str)
        if sum(values.isin(self.categories)) != len(values):
            raise ValueError(
                f"invalid values for `{self.key}`, allowed are: `{self.categories}`"
            )
        if strict:
            possible_categories = self.get_possible_categories(values)
            if len(possible_categories) != len(self.categories):
                raise ValueError(
                    f"Categories {list(set(self.categories)-set(possible_categories))} of feature {self.key} not used. Remove them."
                )
        return values

    def validate_candidental(self, values: pd.Series) -> pd.Series:
        """Method to validate the suggested candidates

        Args:
            values (pd.Series): A dataFrame with candidates

        Raises:
            ValueError: when not all values for a feature are one of the allowed categories

        Returns:
            pd.Series: The passed dataFrame with candidates
        """
        values = values.map(str)
        if sum(values.isin(self.get_allowed_categories())) != len(values):
            raise ValueError(
                f"not all values of input feature `{self.key}` are a valid allowed category from {self.get_allowed_categories()}"
            )
        return values

    def get_forbidden_categories(self):
        """Returns the non-allowed categories

        Returns:
            List[str]: List of the non-allowed categories
        """
        return list(set(self.categories) - set(self.get_allowed_categories()))

    def get_possible_categories(self, values: pd.Series) -> list:
        """Return the superset of categories that have been used in the experimental dataset and
        that can be used in the optimization

        Args:
            values (pd.Series): Series with the values for this feature

        Returns:
            list: list of possible categories
        """
        return sorted(set(list(set(values.tolist())) + self.get_allowed_categories()))

    def to_onehot_encoding(self, values: pd.Series) -> pd.DataFrame:
        """Converts values to a one-hot encoding.

        Args:
            values (pd.Series): Series to be transformed.

        Returns:
            pd.DataFrame: One-hot transformed data frame.
        """
        return pd.DataFrame(
            {f"{self.key}{_CAT_SEP}{c}": values == c for c in self.categories},
            dtype=float,
            index=values.index,
        )

    def from_onehot_encoding(self, values: pd.DataFrame) -> pd.Series:
        """Converts values back from one-hot encoding.

        Args:
            values (pd.DataFrame): One-hot encoded values.

        Raises:
            ValueError: If one-hot columns not present in `values`.

        Returns:
            pd.Series: Series with categorical values.
        """
        cat_cols = [f"{self.key}{_CAT_SEP}{c}" for c in self.categories]
        # we allow here explicitly that the dataframe can have more columns than needed to have it
        # easier in the backtransform.
        if np.any([c not in values.columns for c in cat_cols]):
            raise ValueError(
                f"{self.key}: Column names don't match categorical levels: {values.columns}, {cat_cols}."
            )
        s = values[cat_cols].idxmax(1).str[(len(self.key) + 1) :]
        s.name = self.key
        return s

    def to_dummy_encoding(self, values: pd.Series) -> pd.DataFrame:
        """Converts values to a dummy-hot encoding, dropping the first categorical level.

        Args:
            values (pd.Series): Series to be transformed.

        Returns:
            pd.DataFrame: Dummy-hot transformed data frame.
        """
        return pd.DataFrame(
            {f"{self.key}{_CAT_SEP}{c}": values == c for c in self.categories[1:]},
            dtype=float,
            index=values.index,
        )

    def from_dummy_encoding(self, values: pd.DataFrame) -> pd.Series:
        """Convert points back from dummy encoding.

        Args:
            values (pd.DataFrame): Dummy-hot encoded values.

        Raises:
            ValueError: If one-hot columns not present in `values`.

        Returns:
            pd.Series: Series with categorical values.
        """
        cat_cols = [f"{self.key}{_CAT_SEP}{c}" for c in self.categories]
        # we allow here explicitly that the dataframe can have more columns than needed to have it
        # easier in the backtransform.
        if np.any([c not in values.columns for c in cat_cols[1:]]):
            raise ValueError(
                f"{self.key}: Column names don't match categorical levels: {values.columns}, {cat_cols[1:]}."
            )
        values = values.copy()
        values[cat_cols[0]] = 1 - values[cat_cols[1:]].sum(axis=1)
        s = values[cat_cols].idxmax(1).str[(len(self.key) + 1) :]
        s.name = self.key
        return s

    def to_ordinal_encoding(self, values: pd.Series) -> pd.Series:
        """Converts values to an ordinal integer based encoding.

        Args:
            values (pd.Series): Series to be transformed.

        Returns:
            pd.Series: Ordinal encoded values.
        """
        enc = pd.Series(range(len(self.categories)), index=list(self.categories))
        s = enc[values]
        s.index = values.index
        s.name = self.key
        return s

    def from_ordinal_encoding(self, values: pd.Series) -> pd.Series:
        """Convertes values back from ordinal encoding.

        Args:
            values (pd.Series): Ordinal encoded series.

        Returns:
            pd.Series: Series with categorical values.
        """
        enc = np.array(self.categories)
        return pd.Series(enc[values], index=values.index, name=self.key)

    def sample(self, n: int) -> pd.Series:
        """Draw random samples from the feature.

        Args:
            n (int): number of samples.

        Returns:
            pd.Series: drawn samples.
        """
        return pd.Series(
            name=self.key, data=np.random.choice(self.get_allowed_categories(), n)
        )

    def get_bounds(
        self,
        transform_type: TTransform,
        values: Optional[pd.Series] = None,
    ) -> Tuple[List[float], List[float]]:
        assert isinstance(transform_type, CategoricalEncodingEnum)
        if transform_type == CategoricalEncodingEnum.ORDINAL:
            return [0], [len(self.categories) - 1]
        if transform_type == CategoricalEncodingEnum.ONE_HOT:
            # in the case that values are None, we return the bounds
            # based on the optimization bounds, else we return the true
            # bounds as this is for model fitting.
            if values is None:
                lower = [0.0 for _ in self.categories]
                upper = [
                    1.0 if self.allowed[i] is True else 0.0  # type: ignore
                    for i, _ in enumerate(self.categories)
                ]
            else:
                lower = [0.0 for _ in self.categories]
                upper = [1.0 for _ in self.categories]
            return lower, upper
        if transform_type == CategoricalEncodingEnum.DUMMY:
            lower = [0.0 for _ in range(len(self.categories) - 1)]
            upper = [1.0 for _ in range(len(self.categories) - 1)]
            return lower, upper
        if transform_type == CategoricalEncodingEnum.DESCRIPTOR:
            raise ValueError(
                f"Invalid descriptor transform for categorical {self.key}."
            )
        else:
            raise ValueError(
                f"Invalid transform_type {transform_type} provided for categorical {self.key}."
            )

    def __str__(self) -> str:
        """Returns the number of categories as str

        Returns:
            str: Number of categories
        """
        return f"{len(self.categories)} categories"


class CategoricalOutput(Output):
    type: Literal["CategoricalOutput"] = "CategoricalOutput"
    order_id: ClassVar[int] = 8

    categories: TCategoryVals
    objective: Annotated[
        List[Annotated[float, Field(type=float, ge=0, le=1)]], Field(min_items=2)
    ]

    @validator("categories")
    def validate_categories_unique(cls, categories):
        """validates that categories have unique names

        Args:
            categories (List[str]): List of category names

        Raises:
            ValueError: when categories have non-unique names

        Returns:
            List[str]: List of the categories
        """
        categories = list(categories)
        if len(categories) != len(set(categories)):
            raise ValueError("categories must be unique")
        return categories

    @validator("objective")
    def validate_objective(cls, objective, values):
        if len(objective) != len(values["categories"]):
            raise ValueError("Length of objectives and categories do not match.")
        for o in objective:
            if o > 1:
                raise ValueError("Objective values has to be smaller equal than 1.")
            if o < 0:
                raise ValueError("Objective values has to be larger equal than zero")
        return objective

    def validate_experimental(self, values: pd.Series) -> pd.Series:
        values = values.map(str)
        if sum(values.isin(self.categories)) != len(values):
            raise ValueError(
                f"invalid values for `{self.key}`, allowed are: `{self.categories}`"
            )
        return values

    def to_dict(self) -> Dict:
        """Returns the catergories and corresponding objective values as dictionary"""
        return dict(zip(self.categories, self.objective))

    def __call__(self, values: pd.Series) -> pd.Series:
        return values.map(self.to_dict()).astype(float)



import random
import uuid

import pytest
from pydantic.error_wrappers import ValidationError

from bofire.data_models.base import BaseModel


class Bla(BaseModel):
    a: int = 1


@pytest.fixture
def bla():
    return Bla()


@pytest.fixture
def a():
    return random.randint(10, 100)


@pytest.fixture
def b():
    return str(uuid.uuid4())


def test_assignment_validation(bla, a):
    bla.a = a
    assert bla.a == a


def test_assignment_validation_invalid(bla, b):
    with pytest.raises(ValidationError):
        bla.a = b


def test_forbid_extra():
    with pytest.raises(ValidationError):
        Bla(a=2, mama="papa")



from bofire.surrogates import api as models


def test_surrogate_can_be_loaded_from_data_model(surrogate_spec):
    data_model = surrogate_spec.obj()
    model = models.map(data_model=data_model)
    assert model is not None
